#!/usr/bin/env python3
"""
JGDV, my kitchen sink library.


"""
from importlib.metadata import version
__version__ = version("jgdv")

from ._abstract import protocols as protos
from ._abstract.types import *  # noqa: F403
from ._abstract.error import JGDVError
from ._abstract import prelude
from . import errors
from .decorators import Mixin, Proto

# Subpackage Accessors
from ._abstract import types as Types # noqa: N812
import jgdv.decorators as Decos  # noqa: N812

def identity_fn[T](x:T) -> T:
    """Just returns what it gets

    :param x: A Value
    :returns: The Same Value
    """
    return x



"""
jgdv.cli provides a statemachine based argument parser.

``ParseMachineBase`` defines the state flow,
``ParseMachine`` implements ``__call__`` to start the parsing,
``CLIParserModel`` implements the callbacks for the different states.

``ParamSpec``'s are descriptions of a single argument type,
combined with the parsing logic for that type.
"""
from ._interface import ParamSpec_p, ArgParserModel_p, ParamSource_p, CLIParamProvider_p
from .errors import ParseError
from .parse_machine import ParseMachine
from .parser_model import CLIParserModel
from .param_spec import ParamSpec
from .builder_mixin import ParamSpecMaker_m



"""
TEST File updated

"""
# ruff: noqa: ANN201, ARG001, ANN001, ARG002, ANN202, B011

# Imports
from __future__ import annotations

# ##-- stdlib imports
import logging as logmod
import pathlib as pl
import warnings
# ##-- end stdlib imports

# ##-- 3rd party imports
import pytest
# ##-- end 3rd party imports

##--|
from ..trace_context import TraceContext
##--|

# ##-- types
# isort: off
# General
import abc
import collections.abc
import typing
import types
from typing import cast, assert_type, assert_never
from typing import Generic, NewType, Never
from typing import no_type_check, final, override, overload
# Protocols and Interfaces:
from typing import Protocol, runtime_checkable
# isort: on
# ##-- end types

# ##-- type checking
# isort: off
if typing.TYPE_CHECKING:
    from typing import Final, ClassVar, Any, Self
    from typing import Literal, LiteralString
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv import Maybe
## isort: on
# ##-- end type checking

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:

class TraceExample:

    def start(self) -> None:  # noqa: N802
        blah = 2 + 2
        bloo = 3 + blah
        self._subtestfn(True)  # noqa: FBT003
        self._othertestfn(False)  # noqa: FBT003
        self._subtestfn(bloo > 2)  # noqa: PLR2004

    def _subtestfn(self, val:bool) -> int:  # noqa: FBT001
        amnt : int
        if val:
            amnt = 20
        else:
            amnt = 30

        return amnt

    def _othertestfn(self, val:bool) -> int:
        self._subtestfn(val)
        return 30

# Body:

##--|

class TestTraceContext:

    @pytest.fixture(scope="function")
    def setup(self):
        pass

    ##--|

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_ctor(self):
        match TraceContext(targets=(), track=()):
            case TraceContext():
                assert(True)
            case x:
                assert(False), x

    def test_call_trace(self, caplog):
        expect = [
            "TestTraceContext.test_call_trace ----> TraceExample.start",
            "TraceExample.start   ----> TraceExample._subtestfn",
            "TraceExample.start   ----> TraceExample._othertestfn",
            "TraceExample._othertestfn ----> TraceExample._subtestfn",
            "TraceExample.start   ----> TraceExample._subtestfn",
            "TestTraceContext.test_call_trace ----> TraceContext.__exit__",
        ]
        obj = TraceContext(targets="call",
                           track=("call","trace", "caller")
                           )

        example = TraceExample()
        with obj:
            example.start()

        assert(bool(obj.trace))
        assert(bool(caplog.messages))
        assert(len(caplog.messages) == len(expect))
        for exp,ret in zip(expect, caplog.messages, strict=True):
            assert(exp in ret)

    def test_blacklist(self, caplog):
        expect = [
            "TestTraceContext.test_blacklist ----> TraceExample.start",
            "TraceExample.start   ----> TraceExample._subtestfn",
            "TraceExample.start   ----> TraceExample._othertestfn",
            "TraceExample._othertestfn ----> TraceExample._subtestfn",
            "TraceExample.start   ----> TraceExample._subtestfn",
            # This is Removed:
            # "TestTraceContext.test_blacklist ----> TraceContext.__exit__",
        ]
        obj = TraceContext(targets="call",
                           track=("call","trace", "caller"),
                           )
        obj.blacklist("TraceContext.__exit__")

        example = TraceExample()
        with obj:
            example.start()

        assert(bool(obj.trace))
        assert(bool(caplog.messages))
        assert(len(caplog.messages) == len(expect))
        for exp,ret in zip(expect, caplog.messages, strict=True):
            assert(exp in ret)

    def test_return_trace(self, caplog):
        expect = [
            "TraceExample.start   <---- TraceExample._subtestfn",
            "TraceExample._othertestfn <---- TraceExample._subtestfn",
            "TraceExample.start   <---- TraceExample._othertestfn",
            "TraceExample.start   <---- TraceExample._subtestfn",
            "TestTraceContext.test_return_trace <---- TraceExample.start",
        ]
        obj = TraceContext(targets=("return"),
                           track=("trace",),
                           )

        example = TraceExample()
        with obj:
            example.start()

        assert(bool(obj.trace))
        assert(bool(caplog.messages))
        assert(len(caplog.messages) == len(expect))
        for exp,ret in zip(expect, caplog.messages, strict=True):
            assert(exp in ret)

    def test_line_trace(self, caplog):
        expect = [
            "blah = 2 + 2",
            "bloo = 3 + blah",
            "self._subtestfn(True)  # noqa: FBT003",
            "if val:",
            "amnt = 20",
            "return amnt",
            "self._othertestfn(False)  # noqa: FBT003",
            "self._subtestfn(val)",
            "if val:",
            "amnt = 30",
            "return amnt",
            "return 30",
            "self._subtestfn(bloo > 2)  # noqa: PLR2004",
            "if val:",
            "amnt = 20",
            "return amnt",
            "sys.settrace(None)",
        ]
        obj = TraceContext(targets="line",
                           track=None,
                           )

        example = TraceExample()
        with obj:
            example.start()

        assert(bool(obj.trace))
        assert(bool(caplog.messages))
        assert(len(caplog.messages) == len(expect))
        for exp,ret in zip(expect, caplog.messages, strict=True):
            assert(exp in ret)

    def test_no_logging(self, caplog):
        obj = TraceContext(targets=("call",),
                           track=("trace"),
                           logger=False,
                           )

        example = TraceExample()
        with obj:
            example.start()

        assert(not bool(caplog.messages))
        assert(bool(obj.trace))

class TestTraceContext_writing:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_write_out_file(self, caplog):
        write_target = pl.Path(__file__).with_suffix(".coverage")
        obj = TraceContext(targets=("call", "line", "return"),
                           track=("trace",),
                           logger=False,
                           )

        example = TraceExample()
        with obj:
            example.start()

        obj.write_coverage_file(target=write_target)
        assert(write_target.exists())

    def test_write_out_flat(self, caplog):
        write_target = pl.Path(__file__).parent / "coverage_flat"
        obj = TraceContext(targets=("call", "line", "return"),
                           track=("call", "trace"),
                           logger=False,
                           )

        example = TraceExample()
        with obj:
            example.start()

        obj.write_coverage_dir(root=write_target)
        assert(write_target.exists())
        assert(len(list(write_target.iterdir())) == 2)

    def test_write_out_tree(self, caplog):
        mod_root = pl.Path(__file__).parent.parent.parent
        write_target = pl.Path(__file__).parent / "coverage_tree"
        write_target.mkdir(exist_ok=True)
        obj = TraceContext(targets=("call", "line", "return"),
                           track=("trace",),
                           logger=False,
                           )

        example = TraceExample()
        with obj:
            example.start()

        obj.write_coverage_tree(root=write_target, reroot=mod_root)
        assert(write_target.exists())

    ##--|

    @pytest.mark.skip
    def test_todo(self):
        pass



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import contextlib
import datetime
import enum
import faulthandler
import functools as ftz
import hashlib
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from copy import deepcopy
from time import sleep
from uuid import UUID, uuid1
from weakref import ref

# ##-- end stdlib imports

# ##-- 1st party imports
from .._base import GuardBase
from jgdv.structs.chainguard.mixins.reporter_m import DefaultedReporter_m

# ##-- end 1st party imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
from typing import Any, Never

if TYPE_CHECKING:
    from .._interface import TomlTypes
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, LiteralString
    from typing import Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv import CHECKTYPE
    from .._interface import ChainGuard_i, ProxyWrapper

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

ROOT_INDEX : Final[tuple[str, ...]] = ("<root>", )
##--|
class GuardProxy:
    """ A Base Class for Proxies """
    __slots__ = ("__index", "_data", "_fallback", "_types")
    __index    : tuple[str|int, ...]
    _data      : Maybe[TomlTypes|GuardBase]
    _types     : CHECKTYPE
    _fallback  : Maybe[Any|tuple]

    def __init__(self, data:Maybe[GuardBase|Any], types:Any=None, index:Maybe[Iterable[str|int]]=None, fallback:Maybe[tuple|Any]=()) -> None:  # noqa: ANN401
        self._data      = data
        self._types     = types
        self._fallback  = fallback
        self.__index    = tuple(index or ROOT_INDEX)

    @override
    def __repr__(self) -> str:
        type_str = self._types_str()
        index_str = ".".join(map(str, self._index()))
        return f"<GuardProxy: {index_str}:{type_str}>"

    def __len__(self) -> int:
        if isinstance(self._data, collections.abc.Sized):
            return len(self._data)

        return 0

    def __bool__(self) -> bool:
        return self._data is not None and self._data is not Never

    def __call__(self, *, wrapper:Maybe[ProxyWrapper]=None, **kwargs:Any) -> Any:  # noqa: ANN401
        raise NotImplementedError()

    def __getattr__(self, attr:str) -> GuardProxy:
        raise NotImplementedError()

    def __getitem__(self, keys:int|str|tuple[str]) -> GuardProxy:
        raise NotImplementedError()

    def __contains__(self, other):
        match other:
            case str():
                return other in self._data or other.replace("_","-")
            case x:
                return x in self._data
    ##--|

    def _inject(self, val:Maybe=None, attr:Maybe[int|str|tuple[int|str, ...]]=None, *, clear:bool=False) -> GuardProxy:
        """ create a new proxy that is further into the proxied data """
        match val:
            case _ if clear:
                val = None
            case _:
                pass

        return type(self)(val,
                          types=self._types,
                          index=self._index(attr),
                          fallback=self._fallback)

    def _notify(self) -> None:
        types_str = self._types_str()
        match self._data, self._fallback, self._index():
            case GuardBase(), _, _:
                pass
            case None, (), _:
                pass
            case _, _, []:
                pass
            case None, val, [*index]:
                DefaultedReporter_m.add_defaulted(".".join(map(str, index)),
                                                  val,
                                                  types_str)
            case val, _, [*index]:
                assert(not isinstance(val, GuardBase|None))
                DefaultedReporter_m.add_defaulted(".".join(map(str, index)),
                                                  val,
                                                  types_str)
            case val, flbck, index,:
                msg = "Unexpected Values found: "
                raise TypeError(msg, val, index, flbck)

    def _types_str(self) -> str:
        types_str : str
        match self._types:
            case None:
                return "Any"
            case types.UnionType() as targ:
                types_str = repr(targ)
            case type(__name__=targ):
                types_str = targ
            case _ as targ:
                types_str = str(targ)

        return types_str

    def _match_type(self, val:Maybe) -> Maybe:
        match self._types:
            case type() if not isinstance(val, self._types):
                types_str  = self._types_str()
                index_str  = ".".join(*map(str, self._index(f"({types_str})")))
                msg        = "TomlProxy Value doesn't match declared Type: "
                raise TypeError(msg, index_str, val, self._types)
            case _:
                # TODO handle unions and generics
                pass

        return val

    def _index(self, sub:Maybe[int|str|tuple[int|str, ...]]=None) -> tuple[str|int, ...]:
        """ create a new index from this proxy """
        match sub:
            case None:
                return self.__index
            case int()|str() as x:
                return (*self.__index, x)
            case [*xs]:
                return (*self.__index, *xs)



#!/usr/bin/env python3
"""
Types that help add clarity

Provides a number of type aliases and shorthands.
Such as ``Weak[T]`` for a weakref, ``Stack[T]``, ``Queue[T]`` etc for lists,
and ``Maybe[T]``, ``Result[T, E]``, ``Either[L, R]``.

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import pathlib as pl
import types
from collections import deque
from collections.abc import (Callable, Generator, Hashable, ItemsView,
                             Iterable, Iterator, KeysView, ValuesView)
from re import Match, Pattern
from typing import (Annotated, Any, Final, Literal, Never, Self, TypeGuard,
                    Union, final, Concatenate)
from uuid import UUID, uuid1
from weakref import ref

# ##-- end stdlib imports

# ##-- 3rd party imports
from packaging.specifiers import SpecifierSet
from packaging.version import Version

# ##-- end 3rd party imports

# ##-- Generated Exports
__all__ = ( # noqa: RUF022
# -- Types
"AbsPath", "Builder", "CHECKTYPE", "Char", "CtorFn", "DateTime", "Decorator", "Depth",
"DictItems", "DictKeys", "DictVals", "E_", "Either", "Fifo", "FmtKey", "FmtSpec", "FmtStr",
"Frame", "Func", "Ident", "Lambda", "Lifo", "M_", "Maybe", "MaybeT", "Method", "Module", "Mut",
"NoMut", "Queue", "R_", "RelPath", "Result", "Rx", "RxMatch", "RxStr", "Seconds", "Stack",
"SubOf", "TimeDelta", "Traceback", "Url", "VList", "Vector", "VerSpecStr", "VerStr",
"Weak",

)
# ##-- end Generated Exports

##-- strings
type VerStr                   = Annotated[str, Version] # A Version String
""" A String representing a Version. """
type VerSpecStr               = Annotated[str, SpecifierSet]
""" A String that describes a version set to match. """
type Ident                    = Annotated[str, UUID]
""" A UUID equivalent str. """
type FmtStr                   = Annotated[str, None]
""" Format Strings like 'blah {val} bloo'. """
type FmtSpec                  = Annotated[str, None]
""" Format and conversion parameters. eg: 'blah {val:<9!r}' would be ':<10!r' """
type FmtKey                   = str
""" Names of Keys in a FmtStr or FmtSpec """
type RxStr                    = Annotated[str, Pattern]
""" A String intented to be compiled as a re.Pattern """
type Char                     = Annotated[str, lambda x: len(x) == 1]
""" A String of length 1. A Single character. """
type Url                      = Annotated[str, "url"]
""" A String that is a url. """

##-- end strings

##-- paths
type RelPath = Annotated[pl.Path, lambda x: not x.is_absolute()]
""" A Relative Path. """
type AbsPath = Annotated[pl.Path, lambda x: x.is_absolute()]
""" A Path that is absolute. """

##-- end paths

##-- regex
type Rx       = Pattern
""" A re.Pattern. """
type RxMatch  = Match
""" A re.Match """

##-- end regex

##-- callables
type CtorFn[**I, O]          = Callable[I, O] | type[O]
""" A Callable that constructs a type. """
type Builder[**I, T]         = Callable[I, T]
type Func[**I, O]            = Callable[I, O]
""" A Callable. """
type Method[X:type, **I, O]  = Callable[Concatenate[X, I], O]
""" A Callable that is attached to an object. """
type Decorator[**I1, O1, **I2, O2]       = Func[[Func[I1, O1]], Func[I2, O2]]
""" A Function that decorates a callable. """
type Lambda[**I, O]          = Callable[I, O]
""" A lambda function. """

##-- end callables

##-- containers
type Weak[T]    = ref[T]
""" A Weak reference to a type. """
type Stack[T]   = list[T]
""" A Stack/FIFO. """
type Fifo[T]    = list[T]
""" A Stack/FIFO """
type Queue[T]   = deque[T]
""" A Queue/LIFO """
type Lifo[T]    = list[T]
""" A Queue/LIFO """
type Vector[T,S:int]  = Annotated[list[T], S]
""" A Sized collection of numbers? """

##-- end containers

##-- utils
type VList[T]                = T | list[T]
""" A Value, or a list of values """
type Mut[T]                  = Annotated[T, "Mutable"]
""" A Mutable Value """
type NoMut[T]                = Annotated[T, "Immutable"]
""" A Non-mutable value """

type Maybe[T]                = T | None
""" A Value or None. """
type MaybeT[*I]              = tuple[*I] | None
""" A Tuple of values or None. """
type Result[T, E:Exception]  = T | E
""" A Value, or an exception. """
type Either[L, R]            = L | R
""" One of two different values. """
type SubOf[T]                = TypeGuard[T]
""" A Typeguard for checking for a subtype of T. """

##-- end utils

##-- shorthands
type M_[T]                   = Maybe[T]
""" A Shorthand for Maybe[T] """
type R_[T, E:Exception]      = Result[T,E]
""" A Shorthand for Result[T, E] """
type E_[L, R]            = Either[L,R]
""" A Shorthand for Either[L, R] """

##-- end shorthands

##-- numbers
type Depth      = Annotated[int, lambda x: 0 <= x]
""" A Non-negative int for representing Depth """
type Seconds    = Annotated[int, lambda x: 0 <= x]
""" A Non-negative int for representing Seconds """
type DateTime   = datetime.datetime
""" An alias for datetime.datetime """
type TimeDelta  = datetime.timedelta
""" An alias for datetime.timedelta """

##-- end numbers

##-- dicts
type DictKeys   = KeysView
""" An Alias for the keys of a dict. """
type DictItems  = ItemsView
""" An Alias for the items of a dict. """
type DictVals   = ValuesView
""" An Alias for the values of a dict. """

##-- end dicts

##-- tracebacks and frames
type Traceback = types.TracebackType
""" An Error Traceback """
type Frame     = types.FrameType
""" A Runtime Frame """

##-- end tracebacks and frames

##-- misc
# the stdlib types.UnionType (int | float) is not typing.Union[int, float]
UnionTypes  : Final[types.UnionType]  = types.UnionType | type(Union[int,None])  # noqa: UP007
""" """
type Module                           = types.ModuleType
""" The type of a python module. """

type CHECKTYPE                        = Maybe[type|types.GenericAlias|types.UnionType]
""" """

##-- end misc



#!/usr/bin/env python3
"""
The core implementation of the ChainGuard object,
which is then extended with mixins.
"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
from uuid import UUID, uuid1

# ##-- end stdlib imports

from .errors import GuardedAccessError
from .mixins.access_m import super_get, super_set
from . import _interface as API  # noqa: N812
from jgdv import Proto

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from collections.abc import ItemsView, KeysView, ValuesView
    from ._interface import TomlTypes
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

type InputData           = dict
TABLE_K    : Final[str]  = "__table"
INDEX_K    : Final[str]  = "__index"
MUTABLE_K  : Final[str]  = "__mutable"
ROOT_STR   : Final[str]  = "<root>"
USCORE     : Final[str]  = "_"
DASH       : Final[str]  = "-"
MUTABLE    : Final[str]  = "__mutable"
##--|

class GuardBase(dict):
    """
    Provides access to toml data (ChainGuard.load(apath))
    but as attributes (data.a.path.in.the.data)
    instead of key access (data['a']['path']['in']['the']['data'])

    while also providing typed, guarded access:
    data.on_fail("test", str | int).a.path.that.may.exist()

    while it can then report missing paths:
    data.report_defaulted() -> ['a.path.that.may.exist.<str|int>']
    """

    def __init__(self, data:Maybe[InputData]=None, *, index:Maybe[Iterable[int|str]]=None, mutable:bool=False) -> None:
        super().__init__()
        super_set(self, TABLE_K, data or {})
        super_set(self, INDEX_K, tuple(index or [ROOT_STR]))
        super_set(self, MUTABLE_K, mutable)

    @override
    def __repr__(self) -> str:
        match self._table():
            case dict() as d:
                return f"<{self.__class__.__name__}:{list(d.keys())}>"
            case d:
                return f"<{self.__class__.__name__}:{d}>"

    @override
    def __eq__(self, other:object) -> bool:
        match other:
            case GuardBase() as base:
                return  self._table() == base._table()
            case dict() as adict:
                return self._table() == adict
            case _:
                return False

    @override
    def __hash__(self) -> int: # type: ignore[override]
        return hash(self._table())

    @override
    def __len__(self) -> int:
        return len(self._table())

    def __call__(self) -> None:
        msg = "Don't call a ChainGuard, call a GuardProxy using methods like .on_fail"
        raise GuardedAccessError(msg)

    @override
    def __iter__(self) -> Iterator:
        return iter(getattr(self, TABLE_K).keys())

    @override
    def __contains__(self, _key: object) -> bool:
        match _key:
            case str():
                return _key in self.keys() or _key.replace("_","-") in self.keys()
            case x:
                return x in self.keys()


    @override
    def __setattr__(self, attr:str, value:Any) -> None:
        if not getattr(self, MUTABLE):
            raise TypeError()
        super_set(self, attr, value)

    def __getattr__(self, attr:str) -> Any:  # noqa: ANN401
        return self.__getitem__(attr)

    @override
    def __getitem__(self, keys:int|str|list[str]|tuple[int|str, ...]) -> Any:
        table    : dict
        curr     : dict
        ##--|
        match keys:
            case tuple():
                pass
            case int()|str():
                keys = (keys, )
            case x:
                raise TypeError(type(x))


        table  = self._table()
        curr   = table
        for k in keys:
            match k:
                case str() if k in curr:
                    pass
                case str() if (k:=k.replace(USCORE, DASH)) in curr:
                    pass
                case str():
                    index_s    = ".".join(map(str, self._index(k)))
                    available  = " ".join(table.keys())
                    msg        = f"{index_s} not found, available: [{available}]"
                    raise GuardedAccessError(msg)
                case int() if k < len(curr):
                    pass
                case int():
                    raise GuardedAcccessError("tried to access a list of wrong length")

            match curr.get(k, None):
                case dict() as result:
                    curr = result
                case result:
                    curr = result
        else:
            match curr:
                case dict():
                    return type(self)(curr, index=self._index(keys))
                case [*xs] if all(isinstance(x, dict) for x in xs):
                    index = self._index(keys)
                    return [type(self)(x, index=index) for x in xs]
                case [*xs]:
                    return xs
                case x:
                    return x

    @override
    def get(self, key:str, default:Maybe=None) -> Maybe:
        if key in self:
            return self.__getitem__(key)

        return default
    ##--|
    def _index(self, sub:Maybe[int|str|tuple[int|str, ...]]=None) -> tuple[int|str, ...]:
        match sub:
            case None:
                return super_get(self, INDEX_K)[:]
            case int()|str() as x:
                return (*super_get(self, INDEX_K), x)
            case [*xs]:
                return (*super_get(self, INDEX_K), *xs)
            case x:
                raise TypeError(type(x))

    def _table(self) -> dict:
        return super_get(self, TABLE_K)

    @override
    def keys(self) -> KeysView[str]: # type: ignore[override]
        table = super_get(self, TABLE_K)
        return table.keys()

    @override
    def items(self) -> ItemsView: # type: ignore[override]
        match super_get(self, TABLE_K):
            case dict() as val:
                return val.items()
            case list() as val:
                return {self._index()[-1]: val}.items()
            case GuardBase() as val:
                return val.items()
            case x:
                msg = "Unknown table type"
                raise TypeError(msg, x)

    @override
    def values(self) -> list|ValuesView: # type: ignore[override]
        match super_get(self, TABLE_K):
            case dict() as val:
                return val.values()
            case list() as val:
                return val
            case _:
                raise TypeError()

    @override
    def update(self, *args) -> Never: # type: ignore[override]  # noqa: ANN002
        msg = "ChainGuards are immutable"
        raise NotImplementedError(msg)



"""

"""
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
from collections import ChainMap, defaultdict
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
from statemachine import State, StateMachine
from statemachine.exceptions import TransitionNotAllowed
from statemachine.states import States

# ##-- end 3rd party imports

# ##-- 1st party imports
from jgdv import Proto
from jgdv.structs.chainguard import ChainGuard

# ##-- end 1st party imports

from . import errors
from .param_spec import HelpParam, SeparatorParam, ParamSpec
from . import _interface as API # noqa: N812
from ._interface import ParseResult_d, EXTRA_KEY, EMPTY_CMD, SectionType_e
from ._interface import ParamSpec_p, ParamSpec_i, ArgParserModel_p, ParamSource_p

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
from dataclasses import InitVar, dataclass, field

if TYPE_CHECKING:
    from .param_spec.param_spec import ParamProcessor
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|
# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

HELP            : Final[ParamSpec_i]     = HelpParam()
SEPARATOR       : Final[ParamSpec_i]     = SeparatorParam()


##--|

@Proto(ArgParserModel_p)
class CLIParserModel:
    """

    # {prog} {args} {cmd} {cmd_args}
    # {prog} {args} [{task} {tasks_args}] - implicit do cmd

    """
    SectionTypes  : ClassVar[type[SectionType_e]]  = SectionType_e
    type CmdName                                   = str
    type SubName                                   = str
    type SubConstraint                             = tuple[str,...]
    type Sub_Params                                = tuple[SubConstraint, list[ParamSpec_i]]

    args_initial       : tuple[str, ...]
    args_remaining     : list[str]
    data_cmds          : list[ParseResult_d]
    data_prog          : Maybe[ParseResult_d]
    data_subs          : list[ParseResult_d]
    specs_cmds         : dict[CmdName, list[ParamSpec_i]]
    specs_prog_prefix  : list[str]
    specs_prog         : list[API.ParamSpec_i]
    specs_subs         : dict[SubName, list[ParamSpec_i]]

    implicits          : list[str]
    _subs_constraints  : defaultdict[CmdName, set[SubName]]
    _current_section   : Maybe[tuple[str, list[API.ParamSpec_i]]]
    _current_data      : Maybe[ParseResult_d]
    _separator         : ParamSpec_i
    _help              : ParamSpec_i
    _force_help        : bool
    _report            : Maybe[dict]
    _section_type      : Maybe[SectionType_e]
    _processor         : ParamProcessor

    def __init__(self) -> None:
        self._processor         = ParamSpec._processor
        self._separator         = SEPARATOR
        self._help              = HELP
        self._report            = None
        self._current_section   = None
        self._subs_constraints  = defaultdict(set)
        self._force_help        = False
        self._section_type      = None
        self.args_initial       = ()
        self.args_remaining     = []
        self.args_remaining     = []
        self.data_cmds          = []
        self.data_prog          = None
        self.data_subs          = []
        self.specs_prog_prefix  = ["python"]
        self.specs_prog         = []
        self.specs_cmds         = {}
        self.specs_subs         = {}

    ##--| conditions

    def _has_more_args(self) -> bool:
        return bool(self.args_remaining)
    def _has_no_specs(self) -> bool:
        return not (bool(self.specs_prog)
                    or bool(self.specs_cmds)
                    or bool(self.specs_subs))

    def _has_help_flag_at_tail(self) -> bool:
        return self._processor.matches_head(self._help,
                                            self.args_remaining[-1])

    def _prog_at_front(self) -> bool:
        match self.args_remaining:
            case [head, *_] if any(x in head for x in self.specs_prog_prefix):
                return True
            case _:
                return False

    def _cmd_at_front(self) -> bool:
        match self.args_remaining:
            case [x, *_] if x in self.specs_cmds:
                return True
            case _:
                return False

    def _no_cmd(self) -> bool:
        return not bool(self.data_cmds) and bool(self.implicits)

    def _sub_at_front(self) -> bool:
        match self.args_remaining:
            case [x, *_] if x in self.specs_subs:
                return True
            case _:
                return False

    def _no_sub(self) -> bool:
        return not bool(self.data_subs) and bool(self.implicits)

    def _kwarg_at_front(self) -> bool:
        """ See if theres a kwarg to parse """
        params : list
        if not bool(self.args_remaining):
            return False
        match self._current_section:
            case None:
                return False
            case _, [*params]:
                pass

        head = self.args_remaining[0]
        for param in params:
            match param:
                case API.PositionalParam_p():
                    continue
                case _:
                    if self._processor.matches_head(param, head):
                        return True
        else:
            return False

    def _posarg_at_front(self) -> bool:
        params : list
        if not bool(self.args_remaining):
            return False
        match self._current_section:
            case None:
                return False
            case _, [*params]:
                pass

        head = self.args_remaining[0]
        for param in params:
            match param:
                case API.PositionalParam_p():
                    if self._processor.matches_head(param, head): # type: ignore[arg-type]
                        return True
                case _:
                    continue
        else:
            return False

    def _separator_at_front(self) -> bool:
        if not bool(self.args_remaining):
            return False
        return self._processor.matches_head(self._separator,
                                            self.args_remaining[0])

    ##--| transition actions
    def _insert_implicit_cmd(self) -> None:
        match [(k,v) for k,v in self.implicits.items() if k in self.specs_cmds]:
            case [(str() as x, list() as ys)]:
                logging.debug("Inserting implicit cmd: %s", x)
                self.args_remaining = [*ys, *self.args_remaining]
            case []:
                pass
            case x:
                msg = "Too Many possibly implicit commands"
                raise ValueError(msg, x)

    def _insert_implicit_sub(self) -> None:
        match [(k,v) for k,v in self.implicits.items() if k in self.specs_subs]:
            case [(str() as x, list() as ys)]:
                logging.debug("Inserting implicit sub: %s", x)
                self.args_remaining = [*ys, *self.args_remaining]
            case []:
                pass
            case x:
                msg = "Too Many possibly implicit sub commands"
                raise ValueError(msg, x)

    ##--| state actions

    def prepare_for_parse(self, *, prog:ParamSource_p, cmds:list, subs:list, raw_args:list[str], implicits:Maybe[dict[str, list[str]]]=None) -> None:
        logging.debug("Setting up Parsing : %s", raw_args)
        self.args_initial    = tuple(raw_args[:])
        self.args_remaining  = raw_args[:]
        match implicits:
            case None:
                self.implicits = {}
            case list() as xs:
                self.implicits = {x:[x] for x in xs}
            case dict() as xs:
                self.implicits = xs
            case x:
                raise TypeError(type(x))

        self._prep_prog_lookup(prog)
        self._prep_cmd_lookup(cmds)
        self._prep_sub_lookup(subs)

    def set_force_help(self) -> None:
        match self._help.consume(self.args_remaining[-1:]):
            case dict(), 1:
                self._force_help = True
                self.args_remaining.pop()
            case _:
                pass

    def select_prog_spec(self) -> None:
        logging.debug("Setting Prog Spec")
        self._current_section = ("prog", sorted(self.specs_prog, key=ParamSpec.key_func))
        self._section_type = SectionType_e.prog
        self.args_remaining.pop(0)

    def select_cmd_spec(self) -> None:
        head = self.args_remaining.pop(0)
        logging.debug("Setting Cmd Spec: %s", head)
        match self.specs_cmds.get(head, None):
            case None:
                raise ValueError("No spec found", head)
            case [*params]:
                self._current_section = (head, sorted(params, key=ParamSpec.key_func))
                self._section_type = SectionType_e.cmd

    def select_sub_spec(self) -> None:
        last_cmd     = self.data_cmds[-1].name
        constraints  = self._subs_constraints[last_cmd]
        match self.args_remaining.pop(0):
            case x if x in constraints:
                logging.debug("Setting Sub Spec: %s", x)
                self._current_section  = (x, sorted(self.specs_subs[x], key=ParamSpec.key_func))
                self._section_type     = SectionType_e.sub
            case x:
                msg = "Sub Not Available for cmd"
                raise ValueError(msg, last_cmd, x)


    def initialise_section(self) -> None:
        name      : str
        defaults  : dict
        ##--|
        match self._current_section:
            case str() as name, list() as params:
                logging.debug("Initialising: %s", name)
                defaults = ParamSpec.build_defaults(params)
            case None:
                raise ValueError()
        match self._section_type:
            case SectionType_e.sub:
                last_cmd            = self.data_cmds[-1].name
                self._current_data  = ParseResult_d(name=name, ref=last_cmd, args=defaults)
            case _:
                self._current_data  = ParseResult_d(name=name, args=defaults)

    def parse_kwarg(self) -> None:
        """ try each param until one works """
        logging.debug("Parsing Kwarg")
        params : list[API.ParamSpec_i]
        assert(self._current_data is not None)
        match self._current_section:
            case str(), list() as params:
                pass
            case x:
                raise TypeError(type(x))

        while bool(params):
            if isinstance(params[0], API.PositionalParam_p):
                return
            param = params.pop(0)
            match param.consume(self.args_remaining):
                case None:
                    continue
                case dict() as data, int() as count:
                    self._current_data.args.update(data)
                    self._current_data.non_default.update(data.keys())
                    self.args_remaining = self.args_remaining[count:]
                    return

    def parse_posarg(self) -> None:
        logging.debug("Parsing Posarg")
        params : list[API.ParamSpec_i]
        assert(self._current_data is not None)
        match self._current_section:
            case _, list() as params:
                pass
            case x:
                raise TypeError(type(x))

        while bool(params):
            param = params.pop(0)
            if not isinstance(param, API.PositionalParam_p):
                continue
            match param.consume(self.args_remaining):
                case None:
                    continue
                case dict() as data, int() as count:
                    self._current_data.args.update(data)
                    self._current_data.non_default.update(data.keys())
                    self.args_remaining = self.args_remaining[count:]
                    return

    def parse_separator(self) -> None:
        match self._separator.consume(self.args_remaining):
            case None:
                pass
            case {}, 1:
                self.args_remaining = self.args_remaining[1:]

    def clear_section(self) -> None:
        assert(self._current_data)
        self._current_section = None
        match self._section_type:
            case None:
                raise ValueError()
            case SectionType_e.prog:
                self.data_prog = self._current_data
            case SectionType_e.cmd:
                self.data_cmds.append(self._current_data)
            case SectionType_e.sub:
                self.data_subs.append(self._current_data)
        ##--|
        self._current_data = None

    def cleanup(self) -> None:
        logging.debug("Cleaning up")
        self.args_initial    = ()
        self.args_remaining  = []
        self.specs_cmds      = {}
        self.specs_subs      = {}
    ##--| Report Generation

    def report(self) -> API.ParseReport_d:
        """ Take the parsed results and return a nested dict """
        result : API.ParseReport_d
        cmds : defaultdict[str, list]
        subs : defaultdict[str, list]
        ##--|
        result = API.ParseReport_d(raw=self.args_initial,
                                   remaining=self.args_remaining,
                                   prog=self.data_prog,
                                   _help=self._force_help)
        cmds = defaultdict(list)
        subs = defaultdict(list)

        for cmd in self.data_cmds:
            cmds[cmd.name].append(cmd)
        else:
            result.cmds.update(cmds)

        for sub in self.data_subs:
            assert(sub.ref is not None and sub.ref in result.cmds)
            subs[sub.name].append(sub)
        else:
            result.subs.update(subs)

        # TODO if there were no args, use an empty cmd similar to implicits
        self._report = result
        return result

    ##--| util
    def _prep_prog_lookup(self, prog:ParamSource_p) -> None:
        match prog:
            case ParamSource_p():
                # TODO make it so variable amount of prefix can be consumed
                self.specs_prog_prefix  = [prog.name]
                self.specs_prog         = prog.param_specs()
            case None:
                pass
            case x:
                msg = "Prog needs to be a ParamSource_p"
                raise TypeError(msg, x)

    def _prep_cmd_lookup(self, cmds:list[ParamSource_p]) -> None:
        """ get the param specs for each cmd """
        if not isinstance(cmds, list):
            msg = "cmds needs to be a list"
            raise TypeError(msg, cmds)

        for x in cmds:
            match x:
                case (str() as alias, ParamSource_p() as source):
                    self.specs_cmds[alias] = source.param_specs()
                    self.specs_cmds[source.name] = source.param_specs()
                case ParamSource_p() as source:
                    self.specs_cmds[source.name] = source.param_specs()
                case x:
                    raise TypeError(x)

    def _prep_sub_lookup(self, subs:list[ParamSource_p]) -> None:
        """ for each sub cmd, get it's param specs, but also register the parent cmd constraint """
        if not isinstance(subs, list):
            logging.info("No Subcmd Specs provided for parsing")
            return

        for x in subs:
            match x:
                case [*constraints], ParamSource_p() as source:
                    assert(all(isinstance(c, str) for c in constraints))
                    self.specs_subs[source.name] = source.param_specs()
                    for c in constraints:
                        self._subs_constraints[c].add(source.name)
                case x:
                    raise TypeError(type(x))



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

import copy
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import re
import time
import weakref
from uuid import UUID, uuid1

##-- end builtin imports

from time import sleep
import timeit
from random import random
import jgdv
from jgdv.decorators import MonotonicDec

# ##-- types
# isort: off
# General
import abc
import collections.abc
import typing
import types
from typing import cast, assert_type, assert_never
from typing import Generic, NewType, Never
from typing import no_type_check, final, override, overload
# Protocols and Interfaces:
from typing import Protocol, runtime_checkable
# isort: on
# ##-- end types

# ##-- type checking
# isort: off
if typing.TYPE_CHECKING:
    import pathlib as pl
    from typing import Final, ClassVar, Any, Self
    from typing import Literal, LiteralString
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv import Maybe, Traceback, Func, Method
    from logging import Logger
## isort: on
# ##-- end type checking

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

autorange_fmt : Final[str] = "%-*10s : %-*5d calls took: %-*8.2f seconds"
result_fmt    : Final[str] = "Attempt %-*5d : %-*8.2f seconds"
block_fmt     : Final[str] = "%-*10s : %-*8.2f seconds"
once_fmt      : Final[str] = "%-*10s : %-*8.2f seconds"
##--|
class TimeCtx:
    """ Utility Class to time code execution. """
    _logger      : Maybe[Logger]
    level        : int
    group        : str
    total        : int
    total_ms     : float
    total_s      : float
    _start       : int
    _stop        : int

    def __init__(self, *, logger:Maybe[Logger|Literal[False]]=None, level:int=logmod.INFO, group:Maybe[str]=None) -> None:
        self.level        = level
        self.group : str  = f"{group}::" if group else ""
        self._start       = 0
        self._stop        = 0
        self.total        = 0
        match logger:
            case None:
                self._logger = logging
            case False:
                self._logger = None
            case logmod.Logger() as l:
                self._logger = l

    def __enter__(self) -> Self:
        self._start = time.monotonic_ns()
        return self

    def __exit__(self, etype:Maybe[type], err:Maybe[Exception], tb:Maybe[Traceback]) -> bool:
        self._stop     = time.monotonic_ns()
        self.total     = (self._stop - self._start)
        self.total_ms  = self.total * 0.0001
        self.total_s   = self.total / 1_000_000_000
        if etype:
            return True

        return False

    def _set_name(self, name:Maybe[str]) -> None:
        match name:
            case None:
                pass
            case str() as s:
                self.current_name = self.group + s

    def msg(self, msg:str, *args:Any) -> None:  # noqa: ANN401
        if self._logger is None:
            return

        self._logger.log(self.level, msg, *args)



class TimeDec(MonotonicDec):
    """ Decorate a callable to track its timing """
    _logger  : Maybe[Logger]
    _cache   : Maybe[pl.Path]

    def __init__(self, *, cache:Maybe[pl.Path]=None, logger:Maybe[Logger|Literal[False]]=None, level:Maybe[int]=None, **kwargs:Any) -> None:  # noqa: ANN401
        kwargs.setdefault("mark", "_timetrack_mark")
        kwargs.setdefault("data", "_timetrack_data")
        super().__init__([], **kwargs)
        self._level  = level
        self._cache  = cache
        match logger:
            case logmod.Logger() as l:
                self._logger = l
            case None:
                self._logger = logging
            case False:
                self._logger = None


    @override
    def _wrap_fn_h[**I, O](self, fn:Func[I, O]) -> Func[I, O]:
        logger, level = self._logger, self._level

        def track_time_wrapper(*args:I.args, **kwargs:I.kwargs) -> O:
            with TimeCtx(logger=logger, level=level or logmod.INFO) as timer:
                result = fn(*args, **kwargs)

            timer.msg("Timed: %s took %s seconds", fn.__qualname__, timer.total_s)
            if self._cache:
                now = datetime.datetime.now().strftime("%Y-%m-%d:%H-%M")
                line = f"{now} : {fn.__qualname__} : {timer.total_s}\n"
                with open(self._cache, "a") as f:
                    f.write(line)

            return result

        return track_time_wrapper

    @override
    def _wrap_method_h[**I, O](self, fn:Func[I, O]) -> Func[I, O]:
        return self._wrap_fn_h(fn)



#!/usr/bin/env python3
"""

"""
# ruff: noqa: ANN202, B011
from __future__ import annotations

import logging as logmod
import pathlib as pl
from typing import Final
import warnings

import pytest

from jgdv.structs.locator import JGDVLocator
from ... import DKey, Key_p
from ..path_key import PathDKey

logging = logmod.root

IMP_KEY_BASES               : Final[list[str]]           = ["bob", "bill", "blah", "other", "23boo", "aweg2531", "awe_weg", "aweg-weji-joi"]
EXP_KEY_BASES               : Final[list[str]]           = [f"{{{x}}}" for x in IMP_KEY_BASES]
EXP_P_KEY_BASES             : Final[list[str]]           = ["{bob:wd}", "{bill:w}", "{blah:wi}", "{other:i}"]
PATH_KEYS                   : Final[list[str]]           = ["{bob}/{bill}", "{blah}/{bloo}", "{blah}/{bloo}"]
MUTI_KEYS                   : Final[list[str]]           = ["{bob}_{bill}", "{blah} <> {bloo}", "! {blah}! {bloo}!"]
IMP_IND_KEYS                : Final[list[str]]           = ["bob_", "bill_", "blah_", "other_"]
EXP_IND_KEYS                : Final[list[str]]           = [f"{{{x}}}" for x in IMP_IND_KEYS]

VALID_KEYS                                           = IMP_KEY_BASES + EXP_KEY_BASES + EXP_P_KEY_BASES + IMP_IND_KEYS + EXP_IND_KEYS
VALID_MULTI_KEYS                                     = PATH_KEYS + MUTI_KEYS

class TestPathKey:

    def test_mark(self):
        assert(DKey.MarkOf(PathDKey) is pl.Path)

    def test_expansion(self):
        key = DKey[pl.Path]("test", implicit=True)
        match key.expand({"test":"blah"}):
            case pl.Path() as x:
                assert(x == pl.Path.cwd() / "blah")
                assert(True)
            case x:
                 assert(False), x


    @pytest.mark.xfail
    def test_cwd_expansion(self):
        cwd = pl.Path.cwd()
        key = DKey[pl.Path](".")
        match key.expand({"test":"blah"}):
            case pl.Path() as x:
                assert(x == cwd)
            case x:
                 assert(False), x


    def test_expansion_fail(self):
        key = DKey[pl.Path]("test", implicit=True)
        match key.expand():
            case None:
                assert(True)
            case x:
                 assert(False), x


    def test_loc_expansion(self):
        locs = JGDVLocator(root=pl.Path.cwd())
        locs.update({"test":"blah"})
        key = DKey[pl.Path]("test", implicit=True)
        match key.expand(locs):
            case pl.Path() as x:
                assert(x == pl.Path.cwd() / "blah")
                assert(True)
            case x:
                 assert(False), x


    def test_loc_expansion_miss(self):
        locs = JGDVLocator(root=pl.Path.cwd())
        key = DKey("test", force=PathDKey, implicit=True)
        match key.expand(locs):
            case None:
                assert(True)
            case x:
                 assert(False), x


    @pytest.mark.skip
    def test_todo(self):
        pass



#!/usr/bin/env python3
"""

"""
# ruff: noqa: PLR0133, ANN001, ANN202
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import logging as logmod
import pathlib as pl
import warnings
# ##-- end stdlib imports

# ##-- 3rd party imports
import pytest

# ##-- end 3rd party imports

from .._interface import ArgParserModel_p
from .. import param_spec as Specs  # noqa: N812
from ..parser_model import CLIParserModel
from ..param_spec import ParamSpec
from .. import param_spec as core
from .._interface import ParseResult_d

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType, Never
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
from dataclasses import InitVar, dataclass, field

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

logging = logmod.root

##--| global vars
@pytest.fixture(scope="function")
def model():
    return CLIParserModel()

@pytest.fixture(scope="function")
def PSource(): # noqa: N802

    class ASource:

        def __init__(self, *, name=None, specs=None) -> None:
            self._name = name or "simple"
            self.specs = specs or []

        @property
        def name(self) -> str:
            return self._name

        def param_specs(self) -> list:
            return self.specs

    return ASource


##--|

class TestMachine:

    def test_sanity(self):
        assert(True is not False)

    def test_creation(self, model):
        assert(model is not None)
        assert(isinstance(model, CLIParserModel))
        assert(isinstance(model, ArgParserModel_p))



#!/usr/bin/env python3
"""
Provdes the Main ArgParser_p Protocol,
and the ParseMachineBase StateMachine.

ParseMachineBase descibes the state progression to parse arguments,
while jgdv.cli.arg_parser.CLIParser adds the specific logic to states and transitions
"""
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from collections import ChainMap
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
from statemachine import State, StateMachine, Event
from statemachine.states import States
from statemachine.exceptions import TransitionNotAllowed

# ##-- end 3rd party imports

# ##-- 1st party imports
from jgdv import Maybe
from jgdv.structs.chainguard import ChainGuard

# ##-- end 1st party imports

from . import _interface as API # noqa: N812
from . import errors

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from ._interface import ParamSource_p, ArgParserModel_p
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|
# isort: on
# ##-- end types

logging = logmod.getLogger(__name__)

MAX_STAGES : Final[int] = 200
##--|
class ParseMachine(StateMachine):
    """
    Implemented Parse State Machine

    __call__ with:
    args       : list[str]       -- the cli args to parse (ie: from sys.argv)
    prog : list[ParamSpec_i] -- specs of the top level program
    cmds       : list[ParamSource_p] -- commands that can provide their parameters
    subs    : dict[str, list[ParamSource_p]] -- a mapping from commands -> subcommands that can provide parameters

    A cli call will be of the form:
    {proghead} {prog [kw]args} {cmd} {cmd[kw]args}* [{subs} {subs[kw]args} [-- {subs} {subargs}]* ]? (--help)?

    eg:
    doot -v list -by-group a b c --help
    doot run basic::task -quick --value=2 --help

    Will raise a jgdv.cli.errors.ParseError on failure
    """

    ## States
    # startup states
    Start         = State(initial=True)
    Prepare       = State(enter="prepare_for_parse")
    # Parsing states
    Help         = State(enter="set_force_help")
    Head         = State()
    Section      = State(enter="initialise_section")
    Prog         = State(enter="select_prog_spec")
    Cmd          = State(enter="select_cmd_spec")
    Sub          = State(enter="select_sub_spec")
    Kwargs       = State(enter="parse_kwarg")
    Posargs      = State(enter="parse_posarg")
    Separator    = State(enter="parse_separator")
    Section_end  = State(enter="clear_section")
    # teardown states
    Cleanup  = State(enter="cleanup")
    Report   = State(enter="report")
    End      = State(final=True)

    # Event Transitions
    setup    = (
        Start.to(Prepare)
        | Prepare.to(End,  cond="not _has_more_args")
        | Prepare.to(End,  cond="_has_no_specs")
        | Prepare.to(Help, cond="_has_help_flag_at_tail")
        | Head.from_(Prepare, Help)
    )

    parse = (
        # program / cmd / subs
         Head.to(Prog,   cond="_prog_at_front")
        | Head.to(Cmd,   cond="_cmd_at_front")
        | Head.to(Cmd,   cond="_no_cmd", on="_insert_implicit_cmd")
        | Head.to(Sub,   cond="_sub_at_front")
        | Head.to(Sub,   cond="_no_sub", on="_insert_implicit_sub")
        | Head.to(Report)
        | Section.from_(Prog, Cmd, Sub)
        # args
        | Kwargs.from_(Section, Kwargs, cond="_kwarg_at_front")
        | Posargs.from_(Section, Kwargs, Posargs, cond="_posarg_at_front")
        | Separator.from_(Section, Kwargs, Posargs, cond="_separator_at_front")
        # separator

        # loop
        | Section_end.from_(Section, Kwargs, Posargs, Separator)
        | Section_end.to(Report, cond="not _has_more_args")
        | Section_end.to(Head)
    )

    finish  = (
        Report.to(Cleanup)
        | Cleanup.to(End)
    )

    ##--| composite
    progress = (setup | parse | finish)

    def __init__(self, parser:Maybe[ArgParserModel_p]=None, max:int=MAX_STAGES) -> None:  # noqa: A002
        match parser:
            case API.ArgParserModel_p():
                pass
            case x:
                raise TypeError(type(x))
        super().__init__(parser)
        self.count         = 0
        self.max_attempts  = max

    def __call__(self, args:list[str], *, prog:ParamSource_p, cmds:list[ParamSource_p], subs:list[tuple[tuple[str, ...], ParamSource_p]], implicits:Maybe[dict[str,list[str]]]=None) -> Maybe[dict]:
        assert(self.current_state == self.Start) # type: ignore[has-type]
        while self.current_state != self.End:
            self.progress(prog=prog, cmds=cmds, subs=subs, raw_args=args, implicits=implicits)
        else:
            return self.model._report

    def on_enter_state(self, source:State, event:Event, target:State) -> None:
        logging.debug("Parse(%s, R:%s/%s): %s -> %s -> %s",
                      getattr(self, "count", 0),
                      len(self.model.args_remaining), len(self.model.args_initial),
                      source, event, target)

    def on_exit_state(self) -> None:
        self.count += 1
        if self.max_attempts < self.count:
            logging.debug("Max Parse Stages Occurred")
            raise StopIteration



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
from collections import defaultdict
import linecache
import enum
import functools as ftz
import math
import itertools as itz
import inspect
import logging as logmod
import gc
import re
import sys
import time
import weakref
import trace
from uuid import UUID, uuid1
import pathlib as pl

# ##-- end stdlib imports

# ##-- types
# isort: off
# General
import abc
import collections.abc
import typing
import types
from typing import cast, assert_type, assert_never
from typing import Generic, NewType, Never
from typing import no_type_check, final, override, overload
from typing import Concatenate as Cons
# Protocols and Interfaces:
from typing import Protocol, runtime_checkable
# isort: on
# ##-- end types

# ##-- type checking
# isort: off
if typing.TYPE_CHECKING:
    from ._interface import TraceEvent
    from typing import Final, ClassVar, Any, Self
    from typing import Literal, LiteralString
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv import Maybe, Traceback, Frame
## isort: on
# ##-- end type checking

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

##-- system guards
if not hasattr(sys, "_getframe"):
        msg = "Can't use TraceBuilder on this system, there is no sys._getframe"
        raise ImportError(msg)
##-- end system guards

##--|

class TracebackFactory:
    """ A Helper to simplify access to tracebacks.
    By Default, removes the frames of this tracebuilder from the trace
    ie     : TraceBuilder._get_frames() -> TraceBuilder.__init__() -> call -> call -> root
    will be: call -> call -> root

    use item acccess to limit the frames,
    eg: tb[2:], will remove the two most recent frames from the traceback

    Use as:
    tb = TraceBuilder()
    raise Exception().with_traceback(tb[:])
    """

    def __class_getitem__(cls, item:slice) -> Maybe[Traceback]:
        tbb = cls()
        return tbb[item]

    def __init__(self, *, chop_self:bool=True) -> None:
        self.frames : list[Frame] = []
        self._get_frames()
        if chop_self:
            self.frames = self.frames[2:]

    def __getitem__(self, val:Maybe[slice]=None) -> Maybe[Traceback]:
        match val:
            case None:
                return self.to_tb()
            case slice() | int():
                return self.to_tb(self.frames[val])
            case _:
                msg = "Bad value passed to TraceHelper"
                raise TypeError(msg, val)

    def _get_frames(self) -> None:
        """ from https://stackoverflow.com/questions/27138440
        Builds the frame stack from most recent to least,
        """
        depth = 0
        while True:
            try:
                frame : Frame = sys._getframe(depth)
                depth += 1
            except ValueError:
                break
            else:
                self.frames.append(frame)

    def to_tb(self, frames:Maybe[list[Frame]]=None) -> Maybe[Traceback]:
        top    = None
        frames = frames or self.frames
        for frame in frames:
            top = types.TracebackType(top, frame,
                                      frame.f_lasti,
                                      frame.f_lineno)
        else:
            return top




#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
from collections import defaultdict
import linecache
import enum
import functools as ftz
import math
import itertools as itz
import inspect
import logging as logmod
import gc
import re
import sys
import time
import weakref
import trace
from uuid import UUID, uuid1
import pathlib as pl

# ##-- end stdlib imports

# ##-- types
# isort: off
# General
import abc
import collections.abc
import typing
import types
from typing import cast, assert_type, assert_never
from typing import Generic, NewType, Never
from typing import no_type_check, final, override, overload
from typing import Concatenate as Cons
# Protocols and Interfaces:
from typing import Protocol, runtime_checkable
# isort: on
# ##-- end types

# ##-- type checking
# isort: off
if typing.TYPE_CHECKING:
    from ._interface import TraceEvent
    from typing import Final, ClassVar, Any, Self
    from typing import Literal, LiteralString
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv import Maybe, Traceback, Frame
## isort: on
# ##-- end type checking

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

##-- system guards
if not hasattr(sys, "_getframe"):
        msg = "Can't use TraceBuilder on this system, there is no sys._getframe"
        raise ImportError(msg)
if not hasattr(sys, "settrace"):
    msg = "Cant use a TraceContext on this system, it has no sys.settrace"
    raise ImportError(msg)

##-- end system guards

##--|
DEFAULT_MESSAGES  : Final[dict[str, str]] = {
    "call"        : "----> %s",
    "caller"      : "%-20s ----> %s (l:%s)",
    "return"      : "%-20s <---- %s",
    "line"        : "\t%s:%s : %s",
}

EXEC_LINE      : Final[str]  = r"{}>>>> {}"
NON_EXEC_LINE  : Final[str]  = r"{}     {}"
FIRST_LINE     : Final[str]  = r"{}     {} (NEW FILE: {})"

def must_have_results[T:TraceContext, **I, O](fn:Callable[Cons[T, I],O]) -> Callable[Cons[T, I], O]:
    return fn

    @ftz.wraps
    def _check(self:T, *args:I.args, **kwargs:I.kwargs) -> O:
        assert(self.results)
        return fn(self, *args, **kwargs)

    return _check

class TraceObj:
    __slots__ = ("count", "file", "func", "line_no", "package")
    file     : Maybe[str]
    package  : Maybe[str]
    func     : str
    line_no  : int
    count    : int

    def __init__(self, frame:Frame) -> None:
        self.file     = frame.f_code.co_filename
        self.package  = frame.f_globals.get("__package__", None)
        self.func     = frame.f_code.co_qualname
        self.line_no  = frame.f_lineno
        self.count    = 0
        assert(frame.f_globals.get("__file__", None) == frame.f_code.co_filename)

    @override
    def __repr__(self) -> str:
        return f"<{self.package}:{self.func}:{self.line_no}>"

    @property
    def line(self) -> str:
        assert(self.file)
        return linecache.getline(self.file, self.line_no)

##--|

class TraceWriter:
    exec_line      : str
    non_exec_line  : str
    first_line     : str

    def __init__(self) -> None:
        self.exec_line      = EXEC_LINE
        self.non_exec_line  = NON_EXEC_LINE
        self.first_line     = FIRST_LINE

    def format_trace(self, trace:list[TraceObj]) -> str:  # noqa: F811
        result : list[str] = []
        for obj in trace:
            result.append(obj.line)
        else:
            return "\n".join(result)

    def format_file_execution(self, *, file:str, trace:dict[int, TraceObj], line_nums:bool=False) -> str:  # noqa: F811
        result : list[str]
        num    : str
        ##--|
        # TODO : use a semantic parse to diff executable from non-executable lines
        result = []
        source = linecache.getlines(str(file))
        for i,x in enumerate(source, 1):
            trimmed = x.removesuffix("\n")
            if line_nums:
                num = f"({i}) "
            else:
                num = ""
            match i:
                case 1:
                    result.append(self.first_line.format(num,
                                                         trimmed,
                                                         pl.Path(file).name))
                case int() as potential if potential in trace:
                    # Line executed
                    result.append(self.exec_line.format(num, trimmed))
                case _:
                    # No execution
                    result.append(self.non_exec_line.format(num, trimmed))

        else:
            return "\n".join(result)

##--|

class TraceContext:
    """ Utility to simplify using the trace library, as a context manager

      see https://docs.python.org/3/library/trace.html
    """
    ##--| internal
    _blacklist  : list[str]
    _write_to   : Maybe[pl.Path]
    _logger     : Maybe[logmod.Logger]
    _formatter  : TraceWriter
    _whitelist  : list[str]
    ##--| options
    cache          : Maybe[pl.Path]
    trace_targets  : tuple[TraceEvent, ...]
    track_targets  : tuple[str, ...]
    timestamp      : bool
    log_fmts       : dict[str, str]
    ##--| results
    called         : set[str]
    callers  : defaultdict[str, set[str]]
    counts   : defaultdict[tuple[str, str], int]
    trace    : list[TraceObj]
    lines    : list[TraceObj]

    def __init__(self, *, targets:Maybe[TraceEvent|Iterable[TraceEvent]], track:Maybe[str|Iterable[str]], logger:Maybe[logmod.Logger|Literal[False]]=None, cache:Maybe[pl.Path]=None, timestamp:bool=False, log_fmts:Maybe[dict]=None) -> None:  # noqa: PLR0912, PLR0913
        x   : Any
        xs  : Iterable
        ##--|
        self._blacklist  = [sys.exec_prefix]
        self._whitelist  = []
        self._formatter  = TraceWriter()
        match targets:
            case str() as x:
                self.trace_targets = (cast("TraceEvent", x),)
            case [*xs]:
                self.trace_targets = tuple(xs)
            case None:
                self.trace_targets = ("call",)
            case x:
                raise TypeError(type(x))
        match track:
            case str() as x:
                self.track_targets = (x,)
            case [*xs]:
                self.track_targets = tuple(xs)
            case None:
                self.track_targets = ("trace",)
            case x:
                raise TypeError(type(x))

        assert(all(x in ("call", "line", "return", "exception", "opcode") for x in self.trace_targets))
        self.cache         = cache

        self.timestamp     = timestamp
        self.callers       = defaultdict(set)
        self.called        = set()
        self.counts        = defaultdict(lambda: 0)
        self.trace         = []
        self.log_fmts      = DEFAULT_MESSAGES.copy()
        if log_fmts:
            self.log_fmts.update(log_fmts)

        match logger:
            case False:
                self._logger = None
            case None:
                self._logger = logging
            case logmod.Logger() as log:
                self._logger = log
            case x:
                raise TypeError(type(x))

    def __enter__(self) -> Self:
        sys.settrace(self.sys_trace_h) # type: ignore[arg-type]
        return self

    def __exit__(self, etype:Maybe[type], err:Maybe[Exception], tb:Maybe[Traceback]) -> bool: # type: ignore[exit-return]
        sys.settrace(None)
        return False

    ##--| Filtering

    def blacklist(self, *args:str) -> Self:
        """ Add string's to ignore to the context """
        self._blacklist += args
        return self

    def whitelist(self, *args:str) -> Self:
        self._whitelist += args
        return self

    def ignores(self, curr:Maybe[str|TraceObj]) ->  bool:

        match curr:
            case None:
                return False
            case str() as x if bool(self._whitelist):
                return not any(y in x for y in self._whitelist)
            case str() as x:
                return any(y in x for y in self._blacklist)
            case TraceObj() as obj if bool(self._whitelist):
                return not any(x in self._whitelist for x in [obj.package, obj.file, obj.func])
            case TraceObj() as obj:
                return any(x in self._blacklist for x in [obj.package, obj.file, obj.func])
    ##--| tracer and handlers

    def sys_trace_h(self, frame:Frame, event:TraceEvent, arg:Any) -> Maybe[Callable]:  # noqa: ANN401
        """ The main handler method added to sys for tracing. """
        if self.ignores(frame.f_code.co_qualname):
            return None
        match event:
            case "call":
                self._trace_call(frame)
            case "line":
                self._trace_line(frame)
            case "return":
                self._trace_return(frame)
            case "exception":
                pass
            case "opcode":
                pass
            case x:
                raise TypeError(type(x), x)

        return self.sys_trace_h

    def _trace_call(self, frame:Frame) -> None:
        curr : TraceObj
        ##--|
        if "call" not in self.trace_targets:
            return
        curr = TraceObj(frame)
        # Tracking called functions
        match self._add_called(frame, curr):
            case None:
                self._log("call", curr)
            case TraceObj() as parent:
                self._log("caller", parent.func, curr.func, curr.line_no)

        # Trace
        self._add_trace(curr)

    def _trace_line(self, frame:Frame) -> None:
        if "line" not in self.trace_targets:
            return
        curr = TraceObj(frame)
        self._log("line", curr.package, curr.line_no, curr.line.strip())
        self._add_trace(curr)

    def _trace_return(self, frame:Frame) -> None:
        if "return" not in self.trace_targets:
            return None

        assert(frame.f_back)
        curr    = TraceObj(frame)
        parent  = TraceObj(frame.f_back)
        self._log("return", parent.func, curr.func)
        self._add_trace(curr)

    ##--| assertions

    def assert_called(self, name:str) -> None:
        assert(name in self.called)

    def assert_count(self, package:str, name:str, *, min:Maybe[int]=None, max:Maybe[int]=None) -> None:  # noqa: A002
        assert((package, name) in self.counts)
        match self.counts.get((package, name), None):
            case None:
                raise AssertionError()
            case int() as x:
                assert((min or 0) <= x)
                assert(x < (max or math.inf))

    ##--| IO

    def write_coverage_file(self, *, filter:Maybe[str]=None, target:pl.Path) -> None:  # noqa: A002
        """ Write the coverage trace into a single file
        """
        formatted  : dict[pl.Path, str]
        ##--|
        formatted = self._prepare_trace_for_writing(filter, line_nums=True)
        match target:
            case None:
                pass
            case pl.Path() as f:
                # Write it to file
                joined = "\n".join(formatted.values())
                f.write_text(joined)

    def write_coverage_dir(self, *,  filter:Maybe[str]=None, root:pl.Path) -> None:  # noqa: A002
        """ Write the coverage trace into a flat directory of files
        """
        formatted  : dict[pl.Path, str]
        ##--|
        formatted = self._prepare_trace_for_writing(filter, line_nums=False)
        match root:
            case pl.Path() if not root.exists():
                root.mkdir(parents=True)
                pass
            case pl.Path() if not root.is_dir():
                msg = "Root needs to be a directory"
                raise ValueError(msg)

        for file,text in formatted.items():
            (root / file.name).with_suffix(".coverage").write_text(text)
            pass


    def write_coverage_tree(self, *, filter:Maybe[str]=None, root:pl.Path, reroot:Maybe[pl.Path]=None) -> None:  # noqa: A002
        """ write the coverage trace into a tree of files
        """
        formatted  : dict[pl.Path, str]
        ##--|
        formatted = self._prepare_trace_for_writing(filter, line_nums=False)
        match root:
            case pl.Path() if not root.exists():
                msg = "Root needs to exist"
                raise ValueError(msg)
            case pl.Path() if not root.is_dir():
                msg = "Root needs to be a directory"
                raise ValueError(msg)
            case _:
                pass

        for file,text in formatted.items():
            try:
                if reroot:
                    file = root / file.relative_to(reroot)
                    file.parent.mkdir(parents=True, exist_ok=True)
                elif not file.is_relative_to(root):
                    continue

                file.with_suffix(".coverage").write_text(text)
            except ValueError:
                pass

    def _prepare_trace_for_writing(self, filter:Maybe[str]=None, *, line_nums:bool=False) -> dict[pl.Path, str]:  # noqa: A002, ARG002
        trace      : list[TraceObj]
        grouped    : defaultdict[str, dict[int, TraceObj]]
        formatted  : dict[pl.Path, str]
        ##--|
        # Get the trace
        trace = self.trace
        # filter it
        trace = [x for x in trace if True]
        # Group into files
        grouped = defaultdict(dict)
        for obj in trace:
            assert(obj.file is not None)
            grouped[obj.file][obj.line_no] = obj

        formatted = {}
        for file, _trace in grouped.items():
            # format it
            formatted[pl.Path(file)] = self._formatter.format_file_execution(file=file,
                                                                             trace=_trace,
                                                                             line_nums=line_nums)

        return formatted

    ##--| utils

    def _log(self, key:str, *args) -> None:
        if self._logger is None:
            return

        match self.log_fmts.get(key, None):
            case None:
                return
            case str() as fmt:
                self._logger.info(fmt, *args)

    def _add_trace(self, curr:TraceObj) -> None:
        if "trace" not in self.track_targets:
            return
        self.trace.append(curr)

    def _add_called(self, frame:Frame, curr:TraceObj) -> Maybe[TraceObj]:
        if "call" in self.track_targets:
            assert(curr.package)
            self.called.add(curr.func)
            self.counts[(curr.package, curr.func)] += 1

        if ("caller" in self.track_targets and frame.f_back):
            parent = TraceObj(frame.f_back)
            self.callers[parent.func].add(curr.func)
            return parent

        return None


    def _add_timestamp(self) -> None:
        pass



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import os
import pathlib as pl
import pdb
import re
import signal
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- types
# isort: off
# General
import abc
import collections.abc
import typing
import types
from typing import cast, assert_type, assert_never
from typing import Generic, NewType, Never
from typing import no_type_check, final, override, overload
# Protocols and Interfaces:
from typing import Protocol, runtime_checkable
# isort: on
# ##-- end types

# ##-- type checking
# isort: off
if typing.TYPE_CHECKING:
    from typing import Final, ClassVar, Any, Self
    from typing import Literal, LiteralString
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv import Maybe
## isort: on
# ##-- end type checking

##-- logging
logging    = logmod.getLogger(__name__)
##-- end logging

env        : dict        = cast("dict", os.environ)
PRE_COMMIT : Final[bool] = "PRE_COMMIT" in env
BREAK_HEADER : Final[str] = "\n---- Task Interrupted ---- "

##--| Body:

class SignalHandler:
    """ Install a breakpoint to run on (by default) SIGINT

    disables itself if PRE_COMMIT is in the environment.
    Can act as a context manager

    """

    def __init__(self) -> None:
        self._disabled = PRE_COMMIT

    @staticmethod
    def handle(signum, frame) -> None:
        breakpoint(header=BREAK_HEADER)
        SignalHandler.install()

    @staticmethod
    def install(sig=signal.SIGINT) -> None:
        logging.debug("Installing Basic Interrupt Handler for: %s", signal.strsignal(sig))
        signal.signal(sig, SignalHandler.handle)

    @staticmethod
    def uninstall(sig=signal.SIGINT) -> None:
        logging.debug("Uninstalling Basic Interrupt Handler for: %s", signal.strsignal(sig))
        signal.signal(sig, signal.SIG_DFL)

    def __enter__(self) -> Self:
        if not self._disabled:
            SignalHandler.install()
        return self

    def __exit__(self, etype:Maybe[type], err:Maybe[Exception], tb:Maybe[Traceback]) -> bool:
        if not self._disabled:
            SignalHandler.uninstall()
        if etype is None:
            return False
        return True

class NullHandler:
    """ An interrupt handler that does nothing """

    @staticmethod
    def handle(signum, frame) -> None:
        return

    @staticmethod
    def install(sig=signal.SIGINT) -> None:
        logging.debug("Installing Null Interrupt handler for: %s", signal.strsignal(sig))
        # Install handler for Interrupt signal
        signal.signal(sig, NullHandler.handle)

    @staticmethod
    def uninstall(sig=signal.SIGINT) -> None:
        logging.debug("Uninstalling Null Interrupt handler for: %s", signal.strsignal(sig))
        signal.signal(sig, signal.SIG_DFL)

    def __enter__(self) -> None:
        if not self._disabled:
            NullHandler.install()
        return self

    def __exit__(self, etype:Maybe[type], err:Maybe[Exception], tb:Maybe[Traceback]) -> bool:
        if not self._disabled:
            NullHandler.uninstall()
        if etype is None:
            return False
        return True



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import weakref
from uuid import UUID, uuid1

##-- end builtin imports

# ##-- types
# isort: off
# General
import abc
import collections.abc
import typing
import types
from typing import cast, assert_type, assert_never
from typing import Generic, NewType, Never
from typing import no_type_check, final, override, overload
# Protocols and Interfaces:
from typing import Protocol, runtime_checkable
# isort: on
# ##-- end types

# ##-- type checking
# isort: off
if typing.TYPE_CHECKING:
    from typing import Final, ClassVar, Any, Self
    from typing import Literal, LiteralString
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv import Maybe, Func
## isort: on
# ##-- end type checking

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

from ._interface import DEL_LOG_K
from jgdv.decorators._interface import ClsDecorator_p

DEBUG_DESTRUCT_ON = False

def _log_del(self:Any) -> None:  # noqa: ANN401
    """ standalone del logging """
    logging.warning("Deleting: %s", self)

def _decorate_del(fn:Func[..., None]) -> Func[..., None]:
    """ wraps existing del method """
    @ftz.wraps(fn)
    def _wrapped(self, *args:Any) -> None:  # noqa: ANN001, ANN401
        logging.warning("Deleting: %s", self)
        fn(*args)

    return _wrapped

def LogDel(cls:type) -> type:  # noqa: N802
    """
    A Class Decorator, attaches a debugging statement to the object destructor
    To activate, add classvar of {jgdv.debugging._interface.DEL_LOG_K} = True
    to the class.
    """
    match (getattr(cls, DEL_LOG_K, default=False), # type: ignore[call-overload]
           hasattr(cls, "__del__")):
        case (False, _):
            pass
        case (True, True):
            assert(hasattr(cls, "__del__"))
            setattr(cls, "__del__", _decorate_del(cls.__del__))  # noqa: B010
        case (True, False):
            setattr(cls, "__del__", _log_del)  # noqa: B010
    return cls
##--|
class LogDestruction(ClsDecorator_p):
    """
    A Decorator to log when instances of a class are deleted
    """

    def _debug_del(self) -> None:
        """ standalone del logging """
        logging.warning("Deleting: %s", self)

    def _debug_del_dec(self, fn:Func) -> Callable:
        """ wraps existing del method """

        def _wrapped(_self:object, *args, **kwargs) -> None:
            logging.warning("Deleting: %s", _self)
            fn(_self, *args, **kwargs)

        return _wrapped

    @override
    def __call__[T](self, cls:type[T]) -> type[T]:
        """
        A Class Decorator, attaches a debugging statement to the object destructor
        """
        match (DEBUG_DESTRUCT_ON, hasattr(cls, "__del__")):
            case (False, _):
                pass
            case (True, True):
                setattr(cls, "__del__", self._debug_del_dec(cls.__del__)) # type: ignore[attr-defined]  # noqa: B010
            case (True, False):
                setattr(cls, "__del__", self._debug_del)  # noqa: B010
        return cls



#!/usr/bin/env python3
"""

"""
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import atexit# for @atexit.register
import collections
import contextlib
import datetime
import enum
import faulthandler
import functools as ftz
import hashlib
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
from collections import defaultdict, deque
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref
# ##-- end stdlib imports

# ##-- 3rd party imports
import sh

# ##-- end 3rd party imports

# ##-- 1st party imports
from jgdv import identity_fn, Proto
from jgdv.decorators import MethodMaybe
from jgdv.structs.strang import CodeReference, Strang
from .. import _interface as API # noqa: N812
# ##-- end 1st party imports

from collections.abc import Mapping
from . import _interface as ExpAPI # noqa: N812
from ._interface import Expander_p, SourceChain_d
from ._interface import ExpInst_d, ExpInstChain_d, InstructionFactory_p
from .._interface import Key_p

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, Generic, cast, assert_type, assert_never, Self, Any
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, overload

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Generator
    from collections.abc import Sequence, MutableMapping, Hashable

    from jgdv import Maybe, M_, Func, RxStr, Rx, Ident, FmtStr, CtorFn
    from ._interface import Expandable_p
# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:
type ExpOpts                = ExpAPI.ExpOpts
type InstructionAlts        = list[ExpInst_d]
type InstructionExpansions  = list[ExpInst_d]
type InstructionList        = list[InstructionAlts|ExpInst_d]
DoMaybe                     = MethodMaybe()
# Body:

@Proto(InstructionFactory_p)
class InstructionFactory:
    _ctor : Maybe[type[Key_p]]

    def __init__(self, *, ctor:Maybe[type[Key_p]]=None) -> None:
        self._ctor = ctor

    def set_ctor(self, ctor:Maybe[type[Key_p]]) -> None:
        if ctor is None:
            return
        self._ctor = ctor

    def build_chains(self, val:ExpInst_d, opts:ExpOpts) -> list[ExpInstChain_d|ExpInst_d]:
        match val:
            case ExpInst_d(value=key) if hasattr(key, "exp_generate_chains_h"):
                return cast("list[ExpInstChain_d|ExpInst_d]", val.value.exp_generate_chains_h(val, self, opts))
            case ExpInst_d(value=Key_p() as key):
                chain = [
                    self.build_inst(key, val, opts, decrement=False),
                    self.lift_inst(f"{key:i}", val, opts, decrement=False, implicit=True),
                    self.null_inst(),
                ]
            case ExpInst_d(value=key):
                chain = [
                    self.build_inst(key, val, opts, decrement=False),
                    self.null_inst(),
                ]
            case x:
                raise TypeError(type(x))

        ##--|

        return [ExpInstChain_d(*[x for x in chain if x is not None], root=val.value)]

    def build_inst(self, val:Maybe, root:Maybe[ExpInst_d], opts:ExpOpts, *, decrement:bool=True) -> Maybe[ExpInst_d]:  # noqa: PLR0911
        x          : Any
        implicit   : bool
        lift       : bool
        ##--|
        assert(self._ctor is not None)
        lift, implicit = self._calc_lift(root, opts)
        ##--|
        match val:
            case None:
                return None
            case x if hasattr(x, "exp_to_inst_h"):
                return x.exp_to_inst_h(root, self) # type: ignore[no-any-return]
            case API.NonKey_p() as key:
                return self.literal_inst(key)
            case API.Key_p() as key:
                rec_count  : Maybe[int]  = self._calc_recursion(key, root, opts, decrement=decrement)
                return ExpInst_d(value=key,
                                 convert=key.data.convert,
                                 rec=rec_count if rec_count is not None else key.data.max_expansions,
                                 )
            case x if lift:
                return self.lift_inst(x, root, opts, decrement=decrement, implicit=implicit)
            case pl.Path() | str() as x:
                return self.lift_inst(str(x), root, opts, decrement=decrement, implicit=implicit)
            case x:
                return self.literal_inst(x)

    def literal_inst(self, val:Any) -> ExpInst_d:  # noqa: ANN401
        return ExpInst_d(value=val, literal=True)

    def lift_inst(self, val:str, root:Maybe[ExpInst_d], opts:ExpOpts, *, decrement:bool=False, implicit:bool=False) -> ExpInst_d:
        assert(self._ctor               is               not               None)
        key        : Key_p            = self._ctor(val, implicit=implicit)  # type: ignore[call-arg]
        rec_count  : Maybe[int]       = self._calc_recursion(key, root, opts, decrement=decrement)
        convert    : Maybe[str|bool]  = key.data.convert
        match root:
            case None:
                pass
            case ExpInst_d(convert=convert):
                pass
        return ExpInst_d(value=key,
                         convert=convert,
                         rec=rec_count)

    def null_inst(self) -> ExpInst_d:
        return ExpInst_d(value=None, literal=True)
    ##--|

    def _calc_recursion(self, key:Maybe[Key_p], val:Maybe[ExpInst_d], opts:ExpOpts, *, decrement:bool=True) -> Maybe[int]:
        rec_count : Maybe[int] = None
        match val:
            case _ if "rec" in opts:
                rec_count = opts.get("rec", None)
            case ExpInst_d(rec=int() as rec_count):
                pass
            case ExpInst_d(rec=int() as rec_count):
                pass
            case _:
                pass
        match key:
            case None:
                pass
            case Key_p() as k if rec_count is None:
                rec_count = k.data.max_expansions

        if decrement and rec_count and 0 < rec_count:
            rec_count -= 1

        if val and val.value == key and rec_count is None:
            rec_count = API.RECURSION_GUARD

        return rec_count

    def _calc_lift(self, val:Maybe[ExpInst_d], opts:ExpOpts) -> tuple[bool, bool]:  # noqa: ARG002
        match val:
            case ExpInst_d(value=API.IndirectKey_p(), lift=bool() as lift):
                return lift, True
            case ExpInst_d(lift=bool() as lift):
                return lift, False
            case ExpInst_d(lift=[bool() as lift, bool() as implicit]):
                return lift, implicit
            case _:
                return False, False
##--|

@Proto(Expander_p[API.Key_p])
class DKeyExpanderStack:
    """ A Static class to control expansion.

    In order it does::

        - pre-format the value to (A, coerceA,B, coerceB)
        - (lookup A) or (lookup B) or None
        - manipulates the retrieved value
        - potentially recurses on retrieved values
        - type coerces the value
        - runs a post-coercion hook
        - checks the type of the value to be returned

    During the above, the hooks of Expandable_p will be called on the source,
    if they return nothing, the default hook implementation is used.

    All of those steps are fallible.
    When one of them fails, then the expansion tries to return, in order::

        - a fallback value passed into the expansion call
        - a fallback value stored on construction of the key
        - None

    Redirection Rules::

        - Hit          || {test}  => state[test=>blah]  => blah
        - Soft Miss    || {test}  => state[test_=>blah] => {blah}
        - Hard Miss    || {test}  => state[...]         => fallback or None

    Indirect Keys act as::

        - Indirect Soft Hit ||  {test_}  => state[test_=>blah] => {blah}
        - Indirect Hard Hit ||  {test_}  => state[test=>blah]  => blah
        - Indirect Miss     ||  {test_} => state[...]          => {test_}

    """
    _factory : ClassVar[InstructionFactory] = InstructionFactory()
    _ctor : type[API.Key_p]

    def __init__(self, *, ctor:Maybe[type[API.Key_p]]=None) -> None:
        self._factory.set_ctor(ctor)

    def set_ctor(self, ctor:type[API.Key_p]) -> None:
        """ Dependency injection from DKey.__init_subclass__ """
        self._factory.set_ctor(ctor)

    ##--|

    def redirect(self, source:API.Key_p, *sources:ExpAPI.SourceBases, **kwargs:Any) -> list[Maybe[ExpInst_d]]:  # noqa: ANN401
            return [self.expand(source, *sources, limit=1, **kwargs)]

    def expand(self, key:API.Key_p, *sources:ExpAPI.SourceBases|SourceChain_d, **kwargs:Any) -> Maybe[ExpInst_d]:  # noqa: ANN401, PLR0912, PLR0915
        """ The entry point for expanding a key """
        x : Any
        stack         : list[ExpInst_d|ExpInstChain_d|None]
        result_stack  : list[ExpInst_d|None]

        ##--|
        logging.info("- Expanding: [%s]", repr(key))
        if key.MarkOf(key) is False:
            return ExpInst_d(value=key, literal=True)

        root          = self._factory.build_inst(key, None, {"rec":kwargs.get("limit", None), **kwargs}, decrement=False)
        assert(root is not None)
        stack         = [root]
        result_stack  = []
        source_chain  = SourceChain_d(*sources)
        ## Pop each expansion from the stack,
        ## Adding new expansions back to it,
        ## and literals to the result stack.
        ## Merge instructions trigger result stack entries to be consumed
        while bool(stack):
            logging.info("Stack: %s", stack)
            curr = stack.pop()
            match curr:
                case None | ExpInst_d(value=None):
                    logging.info("[Stack Clear]")
                    stack         = []
                    result_stack  = []
                case ExpInst_d(value=API.NonKey_p()) | ExpInst_d(literal=True) | ExpInst_d(rec=0) as inst:
                    logging.info("[Result shift]: %s", inst)
                    value = self.coerce_result(inst, None, kwargs)
                    result_stack.append(value)
                case ExpInst_d(value=API.Key_p()) as inst:
                    logging.info("[Build Chain]: %s / %s", str(inst.value), result_stack)
                    stack += self._factory.build_chains(inst, kwargs)
                case ExpInstChain_d(merge=int() as count) as inst:
                    logging.info("[Merge Chain]: %s / %s", str(inst.root), result_stack)
                    values, result_stack = reversed(result_stack[-count:]), result_stack[:-count]
                    value         = self.flatten(list(values), inst.root, kwargs)
                    value         = self.coerce_result(value, inst.root, kwargs)
                    self.check_result(value, inst.root, kwargs)
                    result_stack.append(value)
                case ExpInstChain_d() as inst:
                    logging.info("[Lookup Chain]: %s / %s", str(inst.root), result_stack)
                    lookup = self.do_lookup(inst, source_chain, kwargs)
                    stack.append(lookup)
                case x:
                    raise TypeError(type(x))

        else:
            logging.info("Result Stack: %s", result_stack)
            match root.value.data.fallback, kwargs:
                case _, {"fallback": fallback}:
                    pass
                case fallback, _:
                    pass
                case _:
                    fallback = None

            match result_stack:
                case [] | [ExpInst_d(value=None)] | [None]:
                    match fallback:
                        case None:
                            return None
                        case type() as fb_type:
                            return self._factory.literal_inst(fb_type())
                        case fb:
                            return self._factory.literal_inst(fb)
                case [ExpInst_d(value=API.NonKey_p()) as x] | [ExpInst_d(literal=True) as x]:
                    logging.info("|-| %s -> %s", key, x)
                    return self.finalise(x, root.value, kwargs)
                case [ExpInst_d() as x]:
                    msg = "Expansion didn't result in a literal"
                    raise ValueError(msg, x, key)
                case [*_]:
                    msg = "Expansion finished with unmerged results"
                    raise ValueError(msg, result_stack)
                case x:
                    raise TypeError(type(x))
    ##--| Expansion phases

    def do_lookup(self, target:ExpInstChain_d, sources:SourceChain_d, opts:ExpOpts)  -> Maybe[ExpInst_d]:
        """ customisable method for each key subtype
            Target is a list (L1) of lists (L2) of target tuples (T).
            For each L2, the first T that returns a value is added to the final result
            """
        logging.debug("- Lookup: %s", target)
        sources = self.extra_sources(target, sources)
        match sources.lookup(target):
            case None:
                logging.debug("Lookup Failed for: %s", target)
                return None
            case ExpInst_d() as val:
                return val
            case x, y:
                return self._factory.build_inst(x, y, opts)
            case x:
                msg = "Invalid lookup result"
                raise TypeError(msg, x)

    def extra_sources(self, chain:ExpInstChain_d, sources:SourceChain_d) -> SourceChain_d:
        x : Any
        extended : SourceChain_d
        match chain.root:
            case x if not hasattr(x, "exp_extra_sources_h"):
                return sources
            case x:
                extended = x.exp_extra_sources_h(sources)
                assert(extended is not sources)
                return extended

    def flatten(self, values:list[Maybe[ExpInst_d]], root:Key_p, opts:ExpOpts) -> Maybe[ExpInst_d]:
        """
        Flatten separate expansions into a single value
        """
        x : Any
        match values:
            case _ if hasattr(root,"exp_flatten_h"):
                return cast("Maybe[ExpInst_d]", root.exp_flatten_h(values, self._factory, opts))
            case []:
                return None
            case [x, *_]:
                return x
            case x:
                raise TypeError(type(x))

    def coerce_result(self, inst:Maybe[ExpInst_d], root:Maybe[API.Key_p], opts:ExpOpts) -> Maybe[ExpInst_d]:  # noqa: PLR0912
        """
        Coerce the expanded value accoring to source's expansion type ctor
        """
        param         : str
        root_convert  : Maybe[str]        = root.data.convert if root else None
        root_etype    : Maybe[Callable]   = root.data.expansion_type if root else None
        inst_convert  : Maybe[str|bool]   = None
        inst_etype    : Maybe[Callable]   = None
        result        : Maybe[ExpInst_d]  = None
        ##--|
        match inst:
            case None:
                return None
            case ExpInst_d(value=API.Key_p() as k, convert=inst_convert):
                inst_etype = k.data.expansion_type
            case _:
                pass

        if root and hasattr(root, "exp_coerce_h"):
            return cast("Maybe[ExpInst_d]", root.exp_coerce_h(inst, self._factory, opts))

        match (root_etype, root_convert), (inst_etype, inst_convert):
            case _, [_, False]: # Conversion is off
                result = inst
            case [type() as x, None], _: # theres a root expansion type
                if not isinstance(inst.value, x):
                    result = self._factory.literal_inst(x(inst.value))
                else:
                    result = inst
            case [None, _], [type() as x, None]: # theres a inst expansion type
                if not isinstance(inst.value, x):
                    result = self._factory.literal_inst(x(inst.value))
                else:
                    result = inst
            case [_, x], [_, y] if (param:=str(y or x or "")) and bool(param): # conv param
                result = (self._coerce_result_by_conv_param(inst, param, opts)
                          or inst)
            case _:
                result = inst

        ##--|
        logging.debug("- Type Coerced: %r -> %r", root, result)
        result.literal = True
        return result

    def check_result(self, inst:Maybe[ExpInst_d], root:Key_p, opts:ExpOpts) -> None:
        """ check the type of the expansion is correct,
        throw a type error otherwise
        """
        if inst is None:
            return
        if not hasattr(root, "exp_check_result_h"):
            return

        root.exp_check_result_h(inst, opts)

    def finalise(self, inst:ExpInst_d, root:API.Key_p, opts:ExpOpts) -> Maybe[ExpInst_d]:
        """
        A place for any remaining modifications of the result or fallback value
        """
        if hasattr(root, "exp_final_h"):
            return cast("Maybe[ExpInst_d]", root.exp_final_h(inst, root, self._factory, opts))
        else:
            match self.coerce_result(inst, root, opts):
                case None:
                    return None
                case ExpInst_d() as coerced:
                    coerced.literal = True
                    return coerced
                case x:
                    raise TypeError(type(x))

    ##--| Utils

    def _coerce_result_by_conv_param(self, inst:ExpInst_d, conv:str, opts:ExpOpts) -> Maybe[ExpInst_d]:  # noqa: ARG002
        """ really, keys with conv params should been built as a
        specialized registered type, to use an exp_final_hook
        """
        match ExpAPI.EXPANSION_CONVERT_MAPPING.get(conv, None):
            case fn if callable(fn):
                val : Any = fn(inst.value)
                return self._factory.literal_inst(val)
            case None:
                return inst
            case x:
                logging.warning("Unknown Conversion Parameter: %s", x)
                return None



#!/usr/bin/env python3
"""
TEST File updated

"""
# ruff: noqa: ANN202, B011, ANN001, F841, C405

# Imports
from __future__ import annotations

# ##-- stdlib imports
import logging as logmod
import pathlib as pl
import warnings
# ##-- end stdlib imports

# ##-- 3rd party imports
import pytest
# ##-- end 3rd party imports

import random
##--|
from ..malloc_tool import MallocTool
##--|

# ##-- types
# isort: off
# General
import abc
import collections.abc
import typing
import types
from typing import cast, assert_type, assert_never
from typing import Generic, NewType, Never
from typing import no_type_check, final, override, overload
# Protocols and Interfaces:
from typing import Protocol, runtime_checkable
if typing.TYPE_CHECKING:
    from typing import Final, ClassVar, Any, Self
    from typing import Literal, LiteralString
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv import Maybe

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:
EXPECT_NO_FRAMES : Final[str] = """
[TraceMalloc]: --> Entering, tracking 1 frames
[TraceMalloc]: Taking Snapshot: _init_
[TraceMalloc]: Taking Snapshot: before
[TraceMalloc]: Taking Snapshot: after
[TraceMalloc]: Taking Snapshot: cleared
[TraceMalloc]: Taking Snapshot: _final_
[TraceMalloc]: <-- Exited, with 5 snapshots
[TraceMalloc]: ---- Comparing (traceback): before -> after. Objects:
vals = [random.random() for x in range(1000)]
a_dict = {"blah": 23, "bloo": set([1,2,3,4])}
[TraceMalloc]: -- Compare
"""

EXPECT_MULTI_FRAMES : Final[str] = """
[TraceMalloc]: --> Entering, tracking 3 frames
[TraceMalloc]: Taking Snapshot: _init_
[TraceMalloc]: Taking Snapshot: before
[TraceMalloc]: Taking Snapshot: after
[TraceMalloc]: Taking Snapshot: cleared
[TraceMalloc]: Taking Snapshot: _final_
[TraceMalloc]: <-- Exited, with 5 snapshots
[TraceMalloc]: ---- Comparing (traceback): before -> after. Objects:
[TraceMalloc]: -- (obj:0) delta:
[TraceMalloc]: (obj:0, frame: -2) : res = hook_impl.function(*args)                    (_callers.py:
[TraceMalloc]: (obj:0, frame: -1) : result = testfunction(**testargs)                  (python.py:
[TraceMalloc]: (obj:0, frame:  0) : vals = [random.random() for x in range(1000)]
[TraceMalloc]: -- (obj:1) delta:
[TraceMalloc]: (obj:1, frame: -2) : res = hook_impl.function(*args)                    (_callers.py:
[TraceMalloc]: (obj:1, frame: -1) : result = testfunction(**testargs)                  (python.py:
[TraceMalloc]: (obj:1, frame:  0) : a_dict = {"blah": 23, "bloo": set([1,2,3,4])}
[TraceMalloc]: -- Compare

"""

EXPECT_INSPECT : Final[str] = """
[TraceMalloc]: --> Entering, tracking 1 frames
[TraceMalloc]: Taking Snapshot: _init_
[TraceMalloc]: Taking Snapshot: before
[TraceMalloc]: Taking Snapshot: after
[TraceMalloc]: Taking Snapshot: cleared
[TraceMalloc]: Taking Snapshot: _final_
[TraceMalloc]: <-- Exited, with 5 snapshots
[TraceMalloc]: ---- Inspecting: after ----
[TraceMalloc]: (obj:None, frame:  0) : vals = [random.random() for x in range(1000)]
[TraceMalloc]: (obj:None, frame:  0) : a_dict = {"blah": 23, "bloo": set([1,2,3,4])}
[TraceMalloc]: -- inspect --
"""

# Body:

class TestMalloc:

    @pytest.fixture(scope="function")
    def setup(self):
        pass

    ##--|

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_ctor(self):
        match MallocTool():
            case MallocTool():
                assert(True)
            case x:
                assert(False), x

    def test_compare_single_frame(self, caplog):
        expected = [x.strip() for x in EXPECT_NO_FRAMES.splitlines() if bool(x.strip())]
        with MallocTool(frame_count=1) as dm:
            dm.whitelist(__file__)
            dm.blacklist("*.venv")
            val = 2
            dm.snapshot("before")
            vals = [random.random() for x in range(1000)]
            a_dict = {"blah": 23, "bloo": set([1,2,3,4])}
            dm.snapshot("after")
            empty_dict = {"basic": [10, 20]}
            vals = None
            dm.snapshot("cleared")

        dm.compare("before", "after", filter=True, fullpath=False)
        for x in expected:
            assert(x in caplog.text)

    def test_compare_multi_frame(self, caplog):
        expected = [x.strip() for x in EXPECT_MULTI_FRAMES.splitlines() if bool(x.strip())]
        with MallocTool(frame_count=3) as dm:
            dm.whitelist(__file__)
            dm.blacklist("*.venv")
            val = 2
            dm.snapshot("before")
            vals = [random.random() for x in range(1000)]
            a_dict = {"blah": 23, "bloo": set([1,2,3,4])}
            dm.snapshot("after")
            empty_dict = {"basic": [10, 20]}
            vals = None
            dm.snapshot("cleared")

        dm.compare("before", "after", filter=True, fullpath=False)
        for x in expected:
            assert(x in caplog.text)

    def test_basic_inspect(self, caplog):
        expected = [x.strip() for x in EXPECT_INSPECT.splitlines() if bool(x.strip())]
        with MallocTool(frame_count=1) as dm:
            dm.whitelist(__file__)
            dm.blacklist("*.venv")
            val = 2
            dm.snapshot("before")
            vals = [random.random() for x in range(1000)]
            a_dict = {"blah": 23, "bloo": set([1,2,3,4])}
            dm.snapshot("after")
            empty_dict = {"basic": [10, 20]}
            vals = None
            dm.snapshot("cleared")

        dm.inspect("after")
        for x in expected:
            assert(x in caplog.text)



"""
DKey, a str extension for doing things with str format expansion::

    a_key             = DKey("{blah}")
    some_data         = {"blah": "bloo")
    a_key(some_data)  == "bloo"

DKey's manage things such as redirection::

    redirect_key             = DKey("{blah_}")
    some_data                = {"blah_": "bloo", "bloo": "aweg"}
    redirect_key(some_data)  == "aweg"

Also treating input data as a chainmap::

    a_key                  = DKey("{blah}")
    data_1                 = {"not_blah"  : "bloo"}
    data_2                 = {"blah"      : "aweg"}
    a_key(data_1, data_2)  == "aweg"

DKey can also handle multi subkey expansion::

    a_multikey        = DKey("{blah} : {bloo}")
    data              = {"blah": "aweg", "bloo": "qqqq"}
    a_multikey(data)  == "aweg  : qqqq"

"""
from ._interface      import Key_p, DKeyMark_e
from ._util._interface import ExpInst_d
from .errors          import DKeyError
from .dkey            import DKey
from ._util.decorator import DKeyed, DKeyExpansionDecorator

from .keys import SingleDKey, MultiDKey, NonDKey, IndirectDKey

from .special.import_key     import ImportDKey
from .special.args_keys      import ArgsDKey, KwargsDKey
from .special.str_key        import StrDKey
from .special.path_key       import PathDKey



#!/usr/bin/env python3
"""

"""
# ruff: noqa: ARG001, ANN202, N802, ANN001, ANN204, B011, N803, PLR2004

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import logging as logmod
import pathlib as pl
import warnings

# ##-- end stdlib imports

# ##-- 3rd party imports
import pytest

# ##-- end 3rd party imports

from .._interface import ParseResult_d, ParseReport_d
from .. import param_spec as Specs  # noqa: N812
from ..parser_model import CLIParserModel
from ..parse_machine import ParseMachine
from ..param_spec import ParamSpec

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType, Never
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable
    from statemachine import StateMachine

##--|

# isort: on
# ##-- end types

# Logging:
logging = logmod.root

# Global Vars:

@pytest.fixture(scope="function")
def parser(mocker):
    return ParseMachine(CLIParserModel())

@pytest.fixture(scope="function")
def PSource(mocker) -> type:

    class ASource:
        _name : str
        specs : list

        def __init__(self, *, name=None, specs=None):
            self._name = name or "simple"
            self.specs = specs or []

        @property
        def name(self) -> str:
            return self._name

        def param_specs(self) -> list:
            return self.specs

    return ASource

##--| body

class TestMachine:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_creation(self, parser):
        assert(parser is not None)
        assert(isinstance(parser.model, CLIParserModel))

    def test_with_custom_model(self):

        class SubParser(CLIParserModel):
            pass

        parser = ParseMachine(parser=SubParser())
        assert(parser is not None)
        assert(isinstance(parser.model, SubParser))

    def test_empty_parse(self, parser):
        assert(parser.current_state.id == "Start")
        match parser([], prog=None, cmds=[], subs=[]):
            case None:
                assert(parser.current_state_value == "End")
            case x:
                assert(False), x

    def test_parse_no_specs(self, parser):
        assert(parser.current_state.id == "Start")
        match parser(["a", "b","c", "d"], prog=None, cmds=[], subs=[]):
            case None:
                assert(parser.current_state_value == "End")
            case x:
                assert(False), x

    def test_parse_simple_spec(self, parser, PSource):
        assert(parser.current_state.id == "Start")
        prog = PSource(name="blah",
                       specs=[ParamSpec(name="-a")])
        match parser(["blah", "-a", "b","c", "d"], prog=prog, cmds=[], subs=[]):
            case ParseReport_d(remaining=rem, prog=ParseResult_d() as progargs):
                assert(rem == ("b","c","d"))
                assert(progargs.args['a'] is True)
                assert(parser.current_state_value == "End")
            case x:
                assert(False), x

    def test_parse_help(self, parser, PSource):
        assert(parser.current_state.id == "Start")
        prog = PSource(name="bloo",
                       specs=[ParamSpec(name="-a")])
        match parser(["bloo", "--help"], prog=prog, cmds=[], subs=[]):
            case ParseReport_d(remaining=rem, help=True):
                assert(rem == ())
                assert(parser.current_state_value == "End")
            case x:
                assert(False), x

    def test_parse_cmd(self, parser, PSource):
        assert(parser.current_state.id == "Start")
        psource = PSource()
        match parser(["python", psource.name], prog=None, cmds=[psource], subs=[]):
            case ParseReport_d(remaining=(), cmds={"simple":[ParseResult_d(name="simple")]}):
                assert(parser.current_state_value == "End")
            case x:
                assert(False), x

    def test_parse_cmd_with_arg(self, parser, PSource):
        assert(parser.current_state.id == "Start")
        psource = PSource(specs=[ParamSpec[bool](name="-blah", default=False)])
        assert(psource.name == "simple")
        match parser(["python", "simple"],
                     prog=None,
                     cmds=[psource],
                     subs=[]):
            case ParseReport_d(remaining=(), cmds={"simple":[cmdargs]}):
                assert(cmdargs.name == "simple")
                assert(cmdargs.args['blah'] is False)
                assert(parser.current_state_value == "End")
            case x:
                assert(False), x

    def test_parse_cmd_with_parsed_arg(self, parser, PSource):
        assert(parser.current_state.id == "Start")
        psource = PSource(specs=[ParamSpec[bool](name="-blah", default=False)])
        assert(psource.name == "simple")
        match parser(["python", "simple", "-blah"],
                     prog=None,
                     cmds=[psource],
                     subs=[]):
            case ParseReport_d(remaining=(), cmds={"simple":[cmdargs]}):
                assert(cmdargs.name == "simple")
                assert(cmdargs.args['blah'] is True)
                assert(parser.current_state_value == "End")
            case x:
                assert(False), x

    def test_parse_cmd_and_sub(self, parser, PSource):
        assert(parser.current_state.id == "Start")
        ##--|
        prog  = PSource(name="aweg", specs=[])
        cmd   = PSource(name="acmd", specs=[ParamSpec[bool](name="-blah", default=False)])
        sub   = PSource(name="asub", specs=[ParamSpec[bool](name="-blah", default=False)])
        match parser(["aweg", "acmd", "-blah", "asub"],
                     prog=prog,
                     cmds=[cmd],
                     subs=[((cmd.name,), sub)]):
            case ParseReport_d(remaining=(), cmds={"acmd":[cmdargs]}, subs=dict() as subs):
                assert(cmdargs.args['blah'] is True)
                assert("asub" in subs)
                assert(len(subs['asub']) == 1)
                assert(subs['asub'][0].name == "asub")
                assert(subs['asub'][0].args['blah'] is False)
                assert(parser.current_state_value == "End")
            case x:
                assert(False), x

    def test_parse_multi_subs(self, parser, PSource):
        assert(parser.current_state.id == "Start")
        ##--|
        prog  = PSource(name="aweg", specs=[])
        cmd   = PSource(name="acmd", specs=[ParamSpec[bool](name="-blah", default=False)])
        sub   = PSource(name="asub", specs=[
            ParamSpec[bool](name="-blah", default=False),
            ParamSpec[bool](name="-bloo", default=False),
        ])
        match parser(["aweg", "acmd", "-blah", "asub", "--", "asub", "-bloo"],
                     prog=prog,
                     cmds=[cmd],
                     subs=[((cmd.name,), sub)]):
            case ParseReport_d(remaining=(), cmds={"acmd":[cmdargs]}, subs=dict() as subs):
                assert(cmdargs.args['blah'] is True)
                assert("asub" in subs)
                assert(len(subs['asub']) == 2)
                sub1, sub2 = subs['asub']
                assert(sub1.name == "asub")
                assert(sub1.args['bloo'] is False)
                assert(sub1.ref == "acmd")
                assert(sub2.name == "asub")
                assert(sub2.args['bloo'] is True)
                assert(sub2.ref == "acmd")
                assert(parser.current_state_value == "End")
            case x:
                assert(False), x

    def test_multi_cmds(self, parser, PSource):
        assert(parser.current_state.id == "Start")
        ##--|
        prog  = PSource(name="aweg", specs=[])
        cmd1   = PSource(name="acmd", specs=[ParamSpec[bool](name="-blah", default=False)])
        cmd2   = PSource(name="bcmd", specs=[ParamSpec[bool](name="-aweg", default=False)])
        sub   = PSource(name="asub", specs=[
            ParamSpec[bool](name="-blah", default=False),
            ParamSpec[bool](name="-bloo", default=False),
        ])
        match parser(["aweg", "acmd", "-blah", "asub", "--", "bcmd", "-aweg"],
                     prog=prog,
                     cmds=[cmd1, cmd2],
                     subs=[((cmd1.name,), sub)]):
            case ParseReport_d(remaining=(), cmds={"acmd":[acmdargs], "bcmd":[bcmdargs]}, subs=dict() as subs):
                assert(acmdargs.args['blah'] is True)
                assert(bcmdargs.args['aweg'] is True)
                assert("asub" in subs)
                assert(parser.current_state_value == "End")
            case x:
                assert(False), x

    def test_implicit_cmd(self, parser, PSource):
        assert(parser.current_state.id == "Start")
        ##--|
        prog  = PSource(name="aweg", specs=[])
        cmd1  = PSource(name="acmd", specs=[ParamSpec[bool](name="-blah", default=False)])
        cmd2  = PSource(name="bcmd", specs=[ParamSpec[bool](name="-aweg", default=False)])
        sub   = PSource(name="asub", specs=[
            ParamSpec[bool](name="-blah", default=False),
            ParamSpec[bool](name="-bloo", default=False),
        ])
        match parser(["aweg", "asub"],
                     prog=prog,
                     cmds=[cmd1, cmd2],
                     subs=[((cmd1.name,), sub)],
                     implicits=["acmd"],
                     ):
            case ParseReport_d(remaining=(), cmds={"acmd":[acmdargs]}, subs=dict() as subs):
                assert(acmdargs.args['blah'] is False)
                assert("asub" in subs)
                assert(parser.current_state_value == "End")
            case x:
                assert(False), x

    def test_implicit_sub(self, parser, PSource):
        assert(parser.current_state.id == "Start")
        ##--|
        prog  = PSource(name="aweg", specs=[])
        cmd1  = PSource(name="acmd", specs=[ParamSpec[bool](name="-blah", default=False)])
        cmd2  = PSource(name="bcmd", specs=[ParamSpec[bool](name="-aweg", default=False)])
        sub   = PSource(name="asub", specs=[
            ParamSpec[bool](name="-blah", default=False),
            ParamSpec[bool](name="-bloo", default=False),
        ])
        match parser(["aweg", "-bloo"],
                     prog=prog,
                     cmds=[cmd1, cmd2],
                     subs=[((cmd1.name,), sub)],
                     implicits=["acmd", "asub"],
                     ):
            case ParseReport_d(remaining=(), cmds={"acmd":[acmdargs]}, subs=dict() as subs):
                assert(acmdargs.args['blah'] is False)
                assert("asub" in subs)
                assert(subs['asub'][0].args['bloo'] is True)
                assert(parser.current_state_value == "End")
            case x:
                assert(False), x

class TestMachine_Dot:
    """ Write out the dot graphs of the machines """

    @pytest.fixture(scope="function")
    def fsm(self) -> StateMachine:
        return ParseMachine(CLIParserModel())

    @pytest.fixture(scope="function")
    def target(self):
        target = pl.Path(__file__).parent.parent  / "_graphs"
        if not target.exists():
            target.mkdir()
        return target

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_parser_dot(self, fsm, target):
        fsm_name  = type(fsm).__name__
        text      = fsm._graph().to_string()
        tfile     = target / f"_{fsm_name}.dot"
        tfile.write_text(text)
        assert(tfile.exists())



#!/usr/bin/env python3
"""

"""
# ruff: noqa: ANN202, ANN001, ARG002, B011, PLR2004, F841, N802
from __future__ import annotations

import logging as logmod
import pathlib as pl
import warnings

import pytest

from jgdv.structs.locator.errors import DirAbsent, LocationExpansionError, LocationError
from jgdv.structs.locator import JGDVLocator, Location
from jgdv.structs.locator.locator import _LocatorGlobal
from jgdv.structs.dkey import DKey, NonDKey
from .. import _interface as API # noqa: N812

logging = logmod.root
assert(isinstance(JGDVLocator, API.Locator_p))

match JGDVLocator.Current:
    case None:
        initial_loc = JGDVLocator(pl.Path.cwd())
    case x:
        initial_loc = x

@pytest.fixture(scope="function")
def simple() -> JGDVLocator:
    return JGDVLocator(pl.Path.cwd())

@pytest.fixture(scope="function")
def wrap_locs():
    logging.debug("Activating temp locs")
    with JGDVLocator(pl.Path.cwd()) as temp:
        yield temp
##--|

class TestLocator:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_initial(self):
        simple = JGDVLocator(pl.Path.cwd())
        assert(isinstance(simple, JGDVLocator))
        assert(isinstance(simple, API.Locator_p))
        assert(not bool(simple._data))

    def test_update(self, simple):
        assert(not bool(simple._data))
        simple.update({"blah": "file::>bloo"})
        assert(bool(simple._data))
        assert("blah" in simple)

    def test_data_stored_as_locations(self, simple):
        assert(not bool(simple._data))
        simple.update({"blah": "file::>bloo", "aweg":"aweg/abloo"})
        assert(bool(simple._data))
        for x in  simple._data.values():
            assert(isinstance(x, Location))

    def test_registered(self, simple):
        assert(not bool(simple._data))
        simple.update({"a": "file::>blah"})
        assert(bool(simple._data))
        simple.registered("a")

    def test_registered_fail(self, simple):
        assert(not bool(simple._data))
        simple.update({"a": "file::>blah"})
        assert(bool(simple._data))

        with pytest.raises(DirAbsent):
            simple.registered("b")

    def test_update_conflict(self, simple):
        simple.update({"blah": "dir::>bloo"})
        with pytest.raises(LocationError):
            simple.update({"blah": "dir::>blah"})

    def test_update_non_strict(self, simple):
        simple.update({"blah": "dir::>bloo"})
        simple.update({"blah": "dir::>bloo"}, strict=False)

    def test_update_overwrite(self, simple):
        locstr = "dir::>aweg"
        simple = JGDVLocator(pl.Path.cwd())
        simple.update({"blah": "dirr::bloo"})
        simple.update({"blah": "dir::>aweg"}, strict=False)
        assert("blah" in simple)
        assert(simple._data["blah"] == locstr)
        assert(simple.access('blah') == locstr)

    def test_empty_repr(self, simple):
        repr_str = repr(simple)
        assert(repr_str == f"<JGDVLocator (1) : {pl.Path.cwd()!s} : ()>")

    def test_non_empty_repr(self, simple):
        simple.update({"a": "dir::>blah", "b": "dir::>aweg", "awegewag": "dir::>jkwejgio"})
        repr_str = repr(simple)
        assert(repr_str == f"<JGDVLocator (1) : {pl.Path.cwd()!s} : (a, b, awegewag)>")

    def test_clear(self, simple):
        assert(not bool(simple._data))
        simple.update({"a": "dir::>blah"})
        assert("a" in simple)
        simple.clear()
        assert("a" not in simple)

    @pytest.mark.xfail
    def test_metacheck(self, simple):
        assert(False)

class TestLocatorExpansion:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_expansion_attr(self, wrap_locs):
        wrap_locs.update({"todo_bib": "file::>~/github/bibliography/in_progress/todo.bib"})
        match wrap_locs.todo_bib:
            case Location() as x:
                assert(x.path == pl.Path("~/github/bibliography/in_progress/todo.bib"))
                assert(True)
            case x:
                 assert(False), x

    def test_expansion_item(self, wrap_locs):
        wrap_locs.update({"todo_bib": "file::>~/github/bibliography/in_progress/todo.bib"})
        target = pl.Path("~/github/bibliography/in_progress/todo.bib").expanduser().resolve()
        match wrap_locs['{todo_bib}']:
            case pl.Path() as x:
                assert(x == target)
            case x:
                 assert(False), x

    def test_expansion_no_key(self, wrap_locs):
        wrap_locs.update({"todo_bib": "file::>~/github/bibliography/in_progress/todo.bib"})
        target = pl.Path("todo_bib").resolve()
        match wrap_locs['todo_bib']:
            case pl.Path() as x:
                assert(x == target)
            case x:
                 assert(False), x

class TestLocatorUtils:

    def test_normalize(self, simple):
        a_path = pl.Path("a/b/c")
        expected = a_path.absolute()
        result = simple.normalize(a_path)
        assert(result == expected)

    def test_normalize_tilde(self, simple):
        result = simple.normalize(pl.Path("~/blah"))
        assert(result.is_absolute())
        assert(result == pl.Path("~/blah").expanduser())

    def test_normalize_absolute(self, simple):
        result = simple.normalize(pl.Path("/blah"))
        assert(result.is_absolute())
        assert(result == pl.Path("/blah"))

    def test_normalize_relative(self, simple):
        result = simple.normalize(pl.Path("blah"))
        assert(result.is_absolute())
        assert(result == (pl.Path.cwd() / "blah").absolute())

    def test_normalize_relative_with_different_cwd(self):
        simple = JGDVLocator(pl.Path("~/desktop/"))
        result = simple.normalize(pl.Path("blah"))
        assert(result.is_absolute())
        assert(result == (pl.Path("~/desktop/") / "blah").expanduser().absolute())

class TestLocatorGlobal:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_global_Current(self, simple):
        locs = JGDVLocator(pl.Path.cwd())
        assert(isinstance(JGDVLocator.Current, JGDVLocator))

    def test_ctx_manager_basic(self, simple):
        assert(JGDVLocator.Current is initial_loc)
        with JGDVLocator.Current() as locs2:
            assert(JGDVLocator.Current is locs2)

        assert(JGDVLocator.Current is initial_loc)

    @pytest.mark.skip("fails on gh actions")
    def test_ctx_manager_cwd_change(self, simple):
        assert(not bool(simple._data))
        simple.update({"a": "dir::>blah"})
        assert(bool(simple._data))
        assert(simple.root == pl.Path.cwd())
        target = pl.Path("~/Desktop").expanduser().resolve()
        with simple(pl.Path("~/Desktop")) as ctx:
            assert(ctx.root == target)

    @pytest.mark.skip("fails on gh actions")
    def test_stacklen(self, simple):
        assert(_LocatorGlobal.stacklen() == 1)
        locs  = JGDVLocator(pl.Path.cwd())
        assert(_LocatorGlobal.stacklen() == 1)
        with locs() as locs2:
            assert(_LocatorGlobal.stacklen() == 2)
            with locs2() as locs3:
                assert(_LocatorGlobal.stacklen() == 3)

            assert(_LocatorGlobal.stacklen() == 2)

        assert(_LocatorGlobal.stacklen() == 1)
        assert(JGDVLocator.Current is initial_loc)



#/usr/bin/env python3
"""

"""
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import logging as logmod
from copy import deepcopy
from re import Pattern
from uuid import UUID, uuid1
from weakref import ref

# ##-- end stdlib imports

# ##-- 3rd party imports
import pyparsing as pp
import pyparsing.core as ppc

# ##-- end 3rd party imports

# ##-- types
# isort: off
# General
import abc
import collections.abc
import typing
import types
from typing import cast, assert_type, assert_never
from typing import Generic, NewType, Never
from typing import no_type_check, final, override, overload
# Protocols and Interfaces:
from typing import Protocol, runtime_checkable
# isort: on
# ##-- end types

# ##-- type checking
# isort: off
if typing.TYPE_CHECKING:
    from typing import Final, ClassVar, Any, Self
    from typing import Literal, LiteralString
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv import Maybe
## isort: on
# ##-- end type checking

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

MARK                      = ">!<"
MATCHING                  = "Matching" + (" " * 17)
MATCHED                   = "Matched"  + ("-" * 14)
FAILED                    = "Failed"   + ("_" * 11)

try_format   : Final[str] = "{3[0]:>10}{3[1]:3}{3[2]:<10} {0} <{1}> at {2}:  "
match_format : Final[str] = "{4[0]:>10}{4[1]:3}{4[2]:<10}\t {0} <{1}> ({2}) {3}"
fail_format  : Final[str] = "{5[0]:>10}{5[1]:3}{5[2]:<10}\t\t {0} <{1}>: {2} found '{3}' at {4}"

class PyParsingDebuggerControl:

    @staticmethod
    def debug_pyparsing(*flags, all_warnings=False) -> None:
        """ Set pyparsing to debug
        Only applies for parsers created *after* this,
        so has to be set at boot time.
        """
        # pylint: disable=protected-access
        if flags is None:
            flags = [pp.Diagnostics.enable_debug_on_named_expressions]
        if not debug_pyparsing_active_p():
            logging.info("Enabling Debug on %s Parsers", "Named")
            pp.__diag__.enable_debug_on_named_expressions = True
            if bool(flags):
                for flag in flags:
                    pp.enable_diag(flag)

            if all_warnings:
                pp.enable_all_warnings()
            ppc._default_success_debug_action                 = debug_match_action
            ppc._default_start_debug_action                   = debug_try_action
            ppc._default_exception_debug_action               = debug_fail_action
        else:
            logging.warning("PyParsing Debug is already active")

    @staticmethod
    def dfs_activate(*parsers, remove=False):
        """ DFS on a parser, adding debug funcs to named sub parsers """
        # pylint: disable=protected-access
        queue = list(parsers)
        found = set()
        while bool(queue):
            current = queue.pop(0)
            if current is None or id(current) in found:
                continue

            found.add(id(current))

            if bool(current.name) and current.name != current._defaultName and not remove:
                current.set_debug_actions(debug_try_action,
                                        debug_match_action,
                                        debug_fail_action)
            elif remove:
                current.set_debug_actions(None, None, None)

            if hasattr(current, 'expr'):
                queue.append(current.expr)
            elif hasattr(current, 'exprs'):
                queue += current.exprs

    @static
    def debug_active_p() -> bool:
        return pp.__diag__.enable_debug_on_named_expressions

def debug_try_action(instring, loc, expr, *args):
    """
    Log Entry into parsers
    """
    # pylint: disable=unused-argument, consider-using-f-string
    context = _calc_mark_string(instring, loc)
    logging.warning(try_format.format(MATCHING, expr.name, loc, context))

def debug_match_action(instring, startloc, endloc, expr, toks, *args):
    """
    Log Parser Success
    """
    # pylint: disable=unused-argument, consider-using-f-string
    context = _calc_mark_string(instring, endloc)
    logging.warning(match_format.format(MATCHED, expr.name, str(toks), endloc, context))

def debug_fail_action(instring, loc, expr, exc, *args):
    """
    Log Parser failure
    """
    # pylint: disable=unused-argument, consider-using-f-string
    if isinstance(exc, pp.ParseBaseException):
        found_str = exc.pstr[exc.loc:exc.loc + 1].replace(r'\\', '\\').replace("\n", "\\n")
        mark_str  = _calc_mark_string(instring, exc.loc)
        msg       = exc.msg
        loc       = exc.loc
    else:
        found_str = "AssertionError"
        mark_str  = ("", "", "")
        msg       = ""
        loc       = ""

    logging.error(fail_format.format(FAILED, expr.name, msg, found_str, loc, mark_str))

## util

def _calc_mark_string(instring, loc, buffer=10):
    str_len  = len(instring)
    pre_str  = instring[max(0, loc-buffer):max(0, loc)]
    post_str = instring[max(0, loc):min(str_len, loc+buffer)]
    return pre_str.replace("\n", "\\n"), MARK, post_str.replace("\n", "\\n")

# except pp.ParseException as err:
#     import traceback
#     traceback.print_tb(err.__traceback__)
#     logging.warning(f"Parse Failure: {err.markInputline()}")



"""
:class:`JGDVLogConfig<jgdv.logging.config.JGDVLogConfig>` provides a single config point for logging,
both to files and stdout/error. The Log Config is a singleton.

Different loggers can be configured using ``TOML``,
which is parsed into the structure :class:`LoggerSpec<jgdv.logging.logger_spec.LoggerSpec>`.

The module also provides some logging extensions:

- :class:`JGDVLogger<jgdv.logging.logger.JGDVLogger>`
- :class:`JGDVLogRecord<jgdv.logging.record.JGDVLogRecord>`
- :class:`LogCall<jgdv.logging.decorators.LogCall>`, a decorator.

"""

from ._interface import LogLevel_e
from .config import JGDVLogConfig



# adapted from https://stackoverflow.com/questions/50691169
import logging as logmod
import sys
from bdb import Breakpoint
from typing import (Any, Callable, ClassVar, Dict, Generic, Iterable, Iterator,
                    List, Mapping, Match, MutableMapping, Optional, Sequence,
                    Set, Tuple, TypeVar, Union, cast)

logging      = logmod.getLogger(__name__)
trace_logger = logmod.getLogger('jgdv._debug')

class RunningDebugger(metaclass=SingletonMeta):
    """
      The usual debugger stops execution when you call it.
      This starts the tracing, without pausing execution.
      so on a future breakpoint, you can pause and have a trace from where you started the debugger
    """

    def __init__(self):
        super().__init__()
        self.running = False

    def __bool__(self):
        return self.running

    def precmd(self, line):
        trace_logger.info("[db]>>> " + line)
        return line

    def do_help(self, *args):
        print("Acab Debugger Help")
        super().do_help(*args)


    def set_running_trace(self, frame=None):
        """ Start debugging from frame, without pausing execution.
        This is to allow setting a future breakpoint, without having
        to enter the debugger and exit again.
        """
        self.running = True
        if frame is None:
            frame = sys._getframe().f_back
        self.reset()
        while frame:
            frame.f_trace = self.trace_dispatch
            self.botframe = frame
            frame = frame.f_back
        self.set_continue()
        sys.settrace(self.trace_dispatch)

    def set_trace(self, frame=None):
        """Start debugging from frame.

        If frame is not specified, debugging starts from caller's frame.
        """
        self.running = True
        if frame is None:
            frame = sys._getframe().f_back
        # removed "reset" here.
        while frame:
            frame.f_trace = self.trace_dispatch
            self.botframe = frame
            frame = frame.f_back
        self.set_step()
        sys.settrace(self.trace_dispatch)



"""

jgdv.debugging : Utilities for debugging

Provides:

#. SignalHandler     : For installing handlers for interrupts
#. TimeCtx           : A CtxManager for simple timing
#. TimeDec           : A Decorator to time functions
#. TracebackFactory  : For slicing the traceback provided in exceptions
#. TraceContext      : A CtxManager for tracing function calls.
#. MallocTool        : For profiling memory usage.
#. LogDel            : A class decorator for logging when __del__ is called

"""
from .signal_handler import SignalHandler, NullHandler
from .traceback_factory import TracebackFactory
from .traceback_factory import TracebackFactory as TraceBuilder
from .destruction import LogDel



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import fnmatch
import functools as ftz
import itertools as itz
import linecache
import logging as logmod
import pathlib as pl
import re
import time
import tracemalloc
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

import stackprinter
import traceback

# ##-- types
# isort: off
# General
import abc
import collections.abc
import typing
import types
from typing import cast, assert_type, assert_never
from typing import Generic, NewType, Never
from typing import no_type_check, final, override, overload
from typing import Concatenate as Cons
# Protocols and Interfaces:
from typing import Protocol, runtime_checkable
if typing.TYPE_CHECKING:
    from typing import Final, ClassVar, Any, Self
    from typing import Literal, LiteralString
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv import Maybe, Traceback

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

STAT_FORMS       : Final[tuple[str, ...]] = ("traceback", "filename", "lineno")
INIT_SNAP_NAME   : Final[str] = "_init_"
FINAL_SNAP_NAME  : Final[str] = "_final_"

def must_be_started[**I, O](fn:Callable[Cons[MallocTool, I],O]) -> Callable[Cons[MallocTool, I], O]:
    return fn

    @ftz.wraps
    def _check(self:MallocTool, *args:I.args, **kwargs:I.kwargs) -> O:
        assert(self.started)
        return fn(self, *args, **kwargs)

    return _check

##--|

class MallocTool:
    r""" see `tracemalloc <https://docs.python.org/3/library/tracemalloc.html>`_
    in the stdlib. eg:

    ::

        with MallocTool(frame_count=2) as dm:
            dm.whitelist(__file__)
            dm.blacklist("\*tracemalloc.py", all_frames=False)
            val = 2
            dm.snapshot("simple")
            vals = [random.random() for x in range(1000)]
            dm.current()
            dm.snapshot("list")
            vals = None
            dm.current()
            dm.snapshot("cleared")

        dm.compare("simple", "list")
        dm.compare("list", "cleared")
        dm.compare("list", "simple")
        dm.inspect("list")

    """
    frame_count          : int
    started              : bool
    snapshots            : list[tracemalloc.Snapshot]
    named_snapshots      : dict[str, tracemalloc.Snapshot]
    filters              : list[tracemalloc.Filter]

    _logger              : logmod.Logger
    _curr_mem_msg        : str
    _allocation_loc_msg  : str
    _inspect_msg         : str
    _cmp_enter_msg       : str
    _change_msg          : str
    _diff_msg            : str
    _stat_line_msg       : str
    _enter_msg           : str
    _exit_msg            : str
    _take_snap_msg       : str

    def __init__(self, *, frame_count:int=5, logger:Maybe[logmod.Logger]=None) -> None:
        assert(0 < frame_count)
        self._logger              = logger or logging
        self.frame_count          = frame_count
        self.started              = False
        self.snapshots            = []
        self.named_snapshots      = {}
        self.filters              = []
        self.blacklist("*tracemalloc.py", all_frames=False)
        self.blacklist(__file__)
        ##--| Messages:
        self._enter_msg                = "[TraceMalloc]: --> Entering, tracking %s frames"
        self._exit_msg                 = "[TraceMalloc]: <-- Exited, with %s snapshots"
        self._take_snap_msg            = "[TraceMalloc]: Taking Snapshot: %-15s (Current: %-10s, Peak: %s)"
        self._curr_mem_msg             = "[TraceMalloc]: Memory: (Current: %-10s, Peak: %s)"
        self._allocation_loc_msg       = "[TraceMalloc]: Value Allocated At: %s"
        self._inspect_msg              = "[TraceMalloc]: ---- Inspecting: %s ----"
        self._cmp_enter_msg            = "[TraceMalloc]: ---- Comparing (%s): %s -> %s. Objects:%s ----"
        self._gen_exit_msg             = "[TraceMalloc]: -- %s --"
        self._diff_msg                 = "[TraceMalloc]: -- (obj:%s) delta: %s, %s blocks --"
        self._stat_line_msg            = "[TraceMalloc]: (obj:%s, frame:%3s) : %-50s (%s:%s)"
        self._stat_line_no_frames_msg  = "[TraceMalloc]: (obj:%s) %-15s : %-50s (%s:%s)"

    def __enter__(self) -> Self:
        """ Ctx handler to start tracing object allocations """
        self._logger.info(self._enter_msg, self.frame_count)
        tracemalloc.start(self.frame_count)
        self.started = True
        self.snapshot(INIT_SNAP_NAME)
        return self

    @must_be_started
    def __exit__(self, etype:Maybe[type], err:Maybe[Exception], tb:Maybe[Traceback]) -> bool: # type: ignore[exit-return]
        """ Stop tracing allocations """
        self.snapshot(FINAL_SNAP_NAME)
        tracemalloc.stop()
        self.started = False
        self._logger.info(self._exit_msg, len(self.snapshots))
        return False

    ##--| Setup

    def whitelist(self, file_pat:str, *, lineno:Maybe[int]=None, all_frames:bool=True) -> Self:
        """ Add a filter to whitelist a file pattern """
        self.filters.append(
            tracemalloc.Filter(True,  # noqa: FBT003
                               filename_pattern=file_pat,
                               lineno=lineno,
                               all_frames=all_frames),
            )
        return self

    def blacklist(self, file_pat:str, *, lineno:Maybe[int]=None, all_frames:bool=True) -> Self:
        """ Blacklist a file pattern """
        self.filters.append(
            tracemalloc.Filter(False,  # noqa: FBT003
                               filename_pattern=file_pat,
                               lineno=lineno,
                               all_frames=all_frames),
            )
        return self

    ##--| Control

    @must_be_started
    def snapshot(self, name:Maybe[str]=None) -> None:
        """ Take a snapshot of the current memory state """
        traced : Maybe[tuple]
        ##--|
        traced = tracemalloc.get_traced_memory()
        logging.info(self._take_snap_msg, name, self._human(traced[0]), self._human(traced[1]))
        traced = None
        snap = tracemalloc.take_snapshot()
        self.snapshots.append(snap)
        if name and name not in self.named_snapshots:
            self.named_snapshots[name] = snap

        tracemalloc.clear_traces()

    ##--| Report

    @must_be_started
    def current(self, val:Maybe[object]=None) -> None:
        """ Print a brief report about the current memory state """
        traced = tracemalloc.get_traced_memory()
        self._logger.info(self._curr_mem_msg, self._human(traced[0]), self._human(traced[1]))
        if val:
            self._logger.info(self._allocation_loc_msg, tracemalloc.get_object_traceback(val))

    def inspect(self, val:int|str, *, form:str="traceback", filter:bool=True, fullpath:bool=False) -> None:  # noqa: A002
        """ Inspect a single snapshot of the memory state  """
        assert(form in STAT_FORMS)
        self._logger.info(self._inspect_msg, val)
        snap = self._get_snapshot(val, filter=filter)
        for stat in snap.statistics(form):
            self._print_obj_stat_frames(stat, fullpath=fullpath)
        else:
            self._logger.info(self._gen_exit_msg, "inspect")

    def compare(self, val1:int|str, val2:int|str, *, form:str="traceback", filter:bool=True, fullpath:bool=False, count:int=10) -> None:  # noqa: A002, ARG002, PLR0913
        """ Compare two snapshots,
        with control over filtering, output formatting,
        and the number of objects to report about

        """
        differences : list[tracemalloc.StatisticDiff]
        assert(form in STAT_FORMS)
        snap1 = self._get_snapshot(val1, filter=filter)
        snap2 = self._get_snapshot(val2, filter=filter)

        if 1 < self.frame_count:
            printer = self._print_diff_frames
        else:
            printer = self._print_diff_noframes

        differences = snap2.compare_to(snap1, form)
        # TODO differences = self._get_top_n(differences, count=count)
        diff_count  = len(differences)
        self._logger.info(self._cmp_enter_msg, form, val1, val2, diff_count)
        for i, stat in enumerate(differences):
            printer(stat, idx=i, fullpath=fullpath)
        else:
            self._logger.info(self._gen_exit_msg, f"Compare ({diff_count}/{diff_count})")

    ##--| utils

    def _print_diff_noframes(self, stat:tracemalloc.StatisticDiff, *, idx:Maybe[int]=None, fullpath:bool=False) -> None:
        """ Print a diff without showing the stacktrace """
        assert(isinstance(stat, tracemalloc.StatisticDiff))
        tb           = stat.traceback
        frame        = tb[-1]
        size_change  = self._human(stat.size, sign=True)
        if fullpath:
            path = frame.filename
        else:
            path = pl.Path(frame.filename).name
        self._logger.info(self._stat_line_no_frames_msg,
                          idx,
                          size_change,
                          linecache.getline(frame.filename, frame.lineno).strip(),
                          path,
                          frame.lineno,
                          )

    def _print_diff_frames(self, stat:tracemalloc.StatisticDiff, *, idx:Maybe[int]=None, fullpath:bool=False) -> None:
        """ Print a diff, with stacktrace """
        assert(isinstance(stat, tracemalloc.StatisticDiff))
        self._logger.info(self._diff_msg, idx, self._human(stat.size_diff, sign=True), stat.count_diff)
        self._print_obj_stat_frames(stat, idx=idx, fullpath=fullpath)

    def _print_obj_stat_frames(self, stat:tracemalloc.Statistic|tracemalloc.StatisticDiff, *, idx:Maybe[int]=None, fullpath:bool=False) -> None:
        """ Print a stacktrace for a a given object diff """
        assert(isinstance(stat, tracemalloc.Statistic|tracemalloc.StatisticDiff))
        tb     = stat.traceback
        total  = len(tb)-1
        for i, frame in enumerate(tb):
            if fullpath:
                path = frame.filename
            else:
                path = pl.Path(frame.filename).name
            self._logger.info(self._stat_line_msg,
                              idx,
                              i-total,
                              linecache.getline(frame.filename, frame.lineno).strip(),
                              path,
                              frame.lineno,
                              )
        else:
            pass

    def _human(self, num:int, *, sign:bool=False) -> str:
        """ Format a sized number in a human readable way. optionally with a sign prefix """
        return cast("str", tracemalloc._format_size(num, sign)) # type: ignore[attr-defined]

    def _get_snapshot(self, val:int|str, *, filter:bool=True) -> tracemalloc.Snapshot:  # noqa: A002
        """ Retrieve a snapshot,
        with control of whether it is filtered or not
        """
        match val:
            case int() if 0 <= val < len(self.snapshots):
                snap = self.snapshots[val]
            case int() if val < 0:
                snap = self.snapshots[val]
            case str() if val in self.named_snapshots:
                snap = self.named_snapshots[val]
            case _:
                raise TypeError(val)

        if filter:
            return snap.filter_traces(self.filters)

        return snap

    def _check_file_pat(self, file_pat:str, file_name:str) -> bool:
        return fnmatch.fnmatch(file_name, file_pat)

    def _get_top_n(self, stats:list[tracemalloc.StatisticDiff], count:int=10) -> list[tracemalloc.StatisticDiff]:
        r""" Get the top {count} sized objects of a difference """
        raise NotImplementedError()
